/* $OpenLDAP: pkg/ldap/libraries/liblber/sockbuf.c,v 1.54.2.4 2004/04/11 16:58:54 kurt Exp $ */
/* This work is part of OpenLDAP Software <http://www.openldap.org/>.
 *
 * Copyright 1998-2004 The OpenLDAP Foundation.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted only as authorized by the OpenLDAP
 * Public License.
 *
 * A copy of this license is available in the file LICENSE in the
 * top-level directory of the distribution or, alternatively, at
 * <http://www.OpenLDAP.org/license.html>.
 */

#include <legacy/ctype.h>
#include <sk/sk_errno.h>
#include <legacy/string.h>
#include <sk/sk_debug.h>

#ifdef HAVE_IO_H
#include <io.h>
#endif /* HAVE_IO_H */

#if defined( HAVE_FCNTL_H )
#include <fcntl.h>
#endif

#if defined( HAVE_SYS_FILIO_H )
#include <sys/filio.h>
#elif defined( HAVE_SYS_IOCTL_H )
#include <net7g/sys/sk_sockio.h>
#endif

#include "lber-int.h"

#ifndef LBER_MIN_BUFF_SIZE
#define LBER_MIN_BUFF_SIZE		4096
#endif
#ifndef LBER_MAX_BUFF_SIZE
#define LBER_MAX_BUFF_SIZE		65536
#endif
#ifndef LBER_DEFAULT_READAHEAD
#define LBER_DEFAULT_READAHEAD	16384
#endif

#ifdef NOT_USED_BY_NETAPP
Sockbuf *
ber_sockbuf_alloc( void )
{
	Sockbuf			*sb;

	ber_int_options.lbo_valid = LBER_INITIALIZED;

	sb = LBER_CALLOC( 1, sizeof( Sockbuf ) );

	if( sb == NULL ) {
		return NULL;
	}

	ber_int_sb_init( sb );
	return sb;
}

void
ber_sockbuf_free( Sockbuf *sb )
{
	assert( sb != NULL );
	assert( SOCKBUF_VALID( sb ) );

	ber_int_sb_close( sb );
	ber_int_sb_destroy( sb );
	LBER_FREE( sb );
}
#endif	/* NOT_USED_BY_NETAPP */

/* Return values: -1: error, 0: no operation performed or the answer is false,
 * 1: successful operation or the answer is true
 */
int
ber_sockbuf_ctrl( Sockbuf *sb, int opt, void *arg )
{
	Sockbuf_IO_Desc		*p;
	int			ret = 0;

	assert( sb != NULL );

	switch ( opt ) {
		case LBER_SB_OPT_HAS_IO:
			p = sb->sb_iod;
			while ( p && p->sbiod_io != (Sockbuf_IO *)arg ) {
				p = p->sbiod_next;
			}
   
			if ( p ) {
				ret = 1;
			}
			break;

		case LBER_SB_OPT_GET_FD:
			if ( arg != NULL ) {
				*((int *)arg) = sb->sb_fd;
			}
			ret = ( sb->sb_fd == -1 ? -1 : 1);
			break;

		case LBER_SB_OPT_SET_FD:
			sb->sb_fd = *((int *)arg);
			ret = 1;
			break;

#ifdef NOT_USED_BY_NETAPP
		case LBER_SB_OPT_SET_NONBLOCK:
			ret = ber_pvt_socket_set_nonblock( sb->sb_fd, arg != NULL)
				? -1 : 1;
			break;

		case LBER_SB_OPT_DRAIN: {
				/* Drain the data source to enable possible errors (e.g.
				 * TLS) to be propagated to the upper layers
				 */
				char buf[LBER_MIN_BUFF_SIZE];

				do {
					ret = ber_int_sb_read( sb, buf, sizeof( buf ) );
				} while ( ret == sizeof( buf ) );

				ret = 1;
			} break;

		case LBER_SB_OPT_NEEDS_READ:
			ret = ( sb->sb_trans_needs_read ? 1 : 0 );
			break;

		case LBER_SB_OPT_NEEDS_WRITE:
			ret = ( sb->sb_trans_needs_write ? 1 : 0 );
			break;

#endif	/* NOT_USED_BY_NETAPP */

		case LBER_SB_OPT_GET_MAX_INCOMING:
			if ( arg != NULL ) {
				*((int *)arg) = sb->sb_max_incoming;
			}
			ret = 1;
			break;

		case LBER_SB_OPT_SET_MAX_INCOMING:
			sb->sb_max_incoming = *((int *)arg);
			ret = 1;
			break;

		default:
			ret = sb->sb_iod->sbiod_io->sbi_ctrl( sb->sb_iod, opt, arg );
			break;
   }

	return ret;
}

int
ber_sockbuf_add_io( Sockbuf *sb, Sockbuf_IO *sbio, int layer, void *arg )
{
	Sockbuf_IO_Desc		*d, *p, **q;
   
	assert( sb != NULL );
   
	if ( sbio == NULL ) {
		return -1;
	}
   
	q = &sb->sb_iod;
	p = *q;
	while ( p && p->sbiod_level > layer ) {
		q = &p->sbiod_next;
		p = *q;
	}
   
	d = NSLBERI_MALLOC( sizeof( *d ) );
	if ( d == NULL ) {
		return -1;
	}
   
	d->sbiod_level = layer;
	d->sbiod_sb = sb;
	d->sbiod_io = sbio;
	memset( &d->sbiod_pvt, '\0', sizeof( d->sbiod_pvt ) );
	d->sbiod_next = p;
	*q = d;

	if ( sbio->sbi_setup != NULL && ( sbio->sbi_setup( d, arg ) < 0 ) ) {
		return -1;
	}

	return 0;
}
   
int
ber_sockbuf_remove_io( Sockbuf *sb, Sockbuf_IO *sbio, int layer )
{
	Sockbuf_IO_Desc		*p, **q;

	assert( sb != NULL );
	assert( SOCKBUF_VALID( sb ) );
   
	if ( sb->sb_iod == NULL ) {
		return -1;
	}
   
	q = &sb->sb_iod;
	while ( *q != NULL ) {
		p = *q;
		if ( layer == p->sbiod_level && p->sbiod_io == sbio ) {
			if ( p->sbiod_io->sbi_remove != NULL &&
				p->sbiod_io->sbi_remove( p ) < 0 )
			{
				return -1;
			}
			*q = p->sbiod_next;
			NSLBERI_FREE( p );
		break;
		}
		q = &p->sbiod_next;
	}

	return 0;
}

void
ber_pvt_sb_buf_init( Sockbuf_Buf *buf )
{
	buf->buf_base = NULL;
	buf->buf_ptr = 0;
	buf->buf_end = 0;
	buf->buf_size = 0;
}

void
ber_pvt_sb_buf_destroy( Sockbuf_Buf *buf )
{
	assert( buf != NULL);

	if (buf->buf_base) {
		NSLBERI_FREE( buf->buf_base );
	}
	ber_pvt_sb_buf_init( buf );
}

int
ber_pvt_sb_grow_buffer( Sockbuf_Buf *buf, int minsize )
{
	int		pw;
	char			*p;
   
	assert( buf != NULL );

	for ( pw = LBER_MIN_BUFF_SIZE; pw < minsize; pw <<= 1 ) {
		if (pw > LBER_MAX_BUFF_SIZE) {
			return -1;
		}
	}

	if ( buf->buf_size < pw ) {
		p = NSLBERI_REALLOC( buf->buf_base, pw );
		if ( p == NULL ) {
			return -1;
		}
		buf->buf_base = p;
		buf->buf_size = pw;
	}
	return 0;
}

int
ber_pvt_sb_copy_out( Sockbuf_Buf *sbb, char *buf, int len )
{
	int		max;

	assert( buf != NULL );
	assert( sbb != NULL );

	max = sbb->buf_end - sbb->buf_ptr;
	max = ( max < len) ? max : len;
	if ( max ) {
		memcpy( buf, sbb->buf_base + sbb->buf_ptr, max );
		sbb->buf_ptr += max;
		if ( sbb->buf_ptr >= sbb->buf_end ) {
			sbb->buf_ptr = sbb->buf_end = 0;
		}
   }
	return max;
}

int
ber_pvt_sb_do_write( Sockbuf_IO_Desc *sbiod, Sockbuf_Buf *buf_out )
{
	int		to_go;
	int ret;

	assert( sbiod != NULL );
	assert( SOCKBUF_VALID( sbiod->sbiod_sb ) );

	to_go = buf_out->buf_end - buf_out->buf_ptr;
	assert( to_go > 0 );
   
	ret = LBER_SBIOD_WRITE_NEXT( sbiod, buf_out->buf_base +
			buf_out->buf_ptr, to_go );
	if (errno == EINTR) {
		ret = -1;
	}

	if ( ret <= 0 ) {
		return ret;
	}
   
	buf_out->buf_ptr += ret;
	if (buf_out->buf_ptr == buf_out->buf_end) {
		buf_out->buf_end = buf_out->buf_ptr = 0;
	}

	return ret;
}

#ifdef NOT_USED_BY_NETAPP
int
ber_pvt_socket_set_nonblock( ber_socket_t sd, int nb )
{
#if HAVE_FCNTL
	int flags = fcntl( sd, F_GETFL);
	if( nb ) {
		flags |= O_NONBLOCK;
	} else {
		flags &= ~O_NONBLOCK;
	}
	return fcntl( sd, F_SETFL, flags );
		
#elif defined( FIONBIO )
	ioctl_t status = nb ? 1 : 0;
	return ioctl( sd, FIONBIO, &status );
#endif
}
#endif

int
ber_int_sb_init( Sockbuf *sb )
{
	assert( sb != NULL);

	sb->sb_valid=LBER_VALID_SOCKBUF;
	sb->sb_options = 0;
	sb->sb_iod = NULL;
	sb->sb_sd = -1;
#ifdef NOT_USED_BY_NETAPP
	sb->sb_debug = ber_int_debug;
	sb->sb_trans_needs_read = 0;
	sb->sb_trans_needs_write = 0;
#endif
   
	assert( SOCKBUF_VALID( sb ) );
	return 0;
}
   
int
ber_int_sb_close( Sockbuf *sb )
{
	Sockbuf_IO_Desc		*p;

	assert( sb != NULL);
   
	p = sb->sb_iod;
	while ( p ) {
		if ( p->sbiod_io->sbi_close && p->sbiod_io->sbi_close( p ) < 0 ) {
			return -1;
		}
		p = p->sbiod_next;
	}
   
	sb->sb_sd = -1;
   
	return 0;
}

int
ber_int_sb_destroy( Sockbuf *sb )
{
	Sockbuf_IO_Desc		*p;

	assert( sb != NULL);
	assert( SOCKBUF_VALID( sb ) );
   
	while ( sb->sb_iod ) {
		p = sb->sb_iod->sbiod_next;
		ber_sockbuf_remove_io( sb, sb->sb_iod->sbiod_io,
			sb->sb_iod->sbiod_level );
		sb->sb_iod = p;
	}

	return ber_int_sb_init( sb );
}

int
ber_int_sb_read( Sockbuf *sb, void *buf, int len )
{
	int		ret;

	assert( buf != NULL );
	assert( sb != NULL);
	assert( sb->sb_iod != NULL );
	assert( SOCKBUF_VALID( sb ) );

	/* Adding error handling for all the above asserts and
	 * sbi_read being a valid function pointer. We have a 
	 * situation where a giveback is happening and the sbi_read
	 * is being cleared by ldap_cleanup_vars and leads to a panic.
	 * See bug 268659. Check for sbiod_io is more of a defensive
	 * check. We do not know of any cases where this could be
	 * NULL during the normal operation, except when cleaning up
	 * the ldap connection.
	 */
	if (buf == NULL || sb == NULL || sb->sb_iod == NULL ||
			!SOCKBUF_VALID(sb) ||
			sb->sb_iod->sbiod_io == NULL ||
			sb->sb_iod->sbiod_io->sbi_read == NULL) {
		errno = EINTR;
		return -1;
	}
	ret = sb->sb_iod->sbiod_io->sbi_read( sb->sb_iod, buf, len );
	if (errno == EINTR) {
		ret = -1;
	}
	return ret;
}

int
ber_int_sb_write( Sockbuf *sb, void *buf, int len )
{
	int		ret;

	assert( buf != NULL );
	assert( sb != NULL);
	assert( sb->sb_iod != NULL );

	/* Adding error handling for all the above asserts and
	 * sbi_write being a valid function pointer. We have a 
	 * situation where a giveback is happening and the sbi_write
	 * can  be cleared by ldap_cleanup_vars and can lead to a panic.
	 * Check for sbiod_io is more of a defensive
	 * check. We do not know of any cases where this could be
	 * NULL during the normal operation, except when cleaning up
	 * the ldap connection.
	 */
	if (buf == NULL || sb == NULL || sb->sb_iod == NULL ||
			!SOCKBUF_VALID(sb) ||
			sb->sb_iod->sbiod_io == NULL ||
			sb->sb_iod->sbiod_io->sbi_write == NULL) {
		errno = EINTR;
		return -1;
	}
	ret = sb->sb_iod->sbiod_io->sbi_write( sb->sb_iod, buf, len );
	if (errno == EINTR) {
		ret = -1;
	}
	return ret;
}
