/* $OpenLDAP: pkg/ldap/libraries/libldap/cyrus.c,v 1.88.2.12 2004/04/12 15:17:41 kurt Exp $ */
/* This work is part of OpenLDAP Software <http://www.openldap.org/>.
 *
 * Copyright 1998-2004 The OpenLDAP Foundation.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted only as authorized by the OpenLDAP
 * Public License.
 *
 * A copy of this license is available in the file LICENSE in the
 * top-level directory of the distribution or, alternatively, at
 * <http://www.OpenLDAP.org/license.html>.
 */


#include <legacy/string.h>
#include <legacy/time.h>
#include <sk/sk_errno.h>
#include <legacy/ctype.h>
#include <sys/limits.h>
#include <sk/sk_debug.h>

#include <ldap.h>
#include "ldap-int.h"

#include <sys/types.h>
#include <sasl.h>
#include <ldap/ontapldap/ldap_vfiler.h>

/*
 * Malloc tags definitions
 */
MALLOC_DECLARE(M_LDAP);


#define SASL_CONST const

/* LBER IO operations stacking levels */
#define LBER_SBIOD_LEVEL_PROVIDER       10
#define LBER_SBIOD_LEVEL_TRANSPORT      20
#define LBER_SBIOD_LEVEL_APPLICATION    30

#define LDAP_PVT_SASL_LOCAL_SSF 71      /* SSF for Unix Domain Socke
ts */

extern char *nsldapi_host_connected_to(Sockbuf *sb);

sk_mutex_t*     ldap_pvt_sasl_mutex_new(void);
int             ldap_pvt_sasl_mutex_lock(sk_mutex_t *mutex);
int             ldap_pvt_sasl_mutex_unlock(sk_mutex_t *mutex);
void            ldap_pvt_sasl_mutex_dispose(sk_mutex_t *mutex);

/*
* Various Cyrus SASL related stuff.
*/

static const sasl_callback_t client_callbacks[] = {
	{ SASL_CB_USER, NULL, NULL },
	{ SASL_CB_AUTHNAME, NULL, NULL },
	{ SASL_CB_PASS, NULL, NULL },
	{ SASL_CB_ECHOPROMPT, NULL, NULL },
	{ SASL_CB_NOECHOPROMPT, NULL, NULL },
	{ SASL_CB_LIST_END, NULL, NULL }
};

int ldap_int_sasl_init( void )
{
	/* XXX not threadsafe */
	if ( VV_sasl_initialized ) {
		return 0;
	}

	sasl_set_mutex(
		ldap_pvt_sasl_mutex_new,
		ldap_pvt_sasl_mutex_lock,
		ldap_pvt_sasl_mutex_unlock,    
		ldap_pvt_sasl_mutex_dispose );    

	if ( sasl_client_init( NULL ) == SASL_OK ) {
		VV_sasl_initialized = 1;
		return 0;
	}

	return -1;
}

/*
 * SASL encryption support for LBER Sockbufs
 */

struct sb_sasl_data {
	sasl_conn_t		*sasl_context;
	unsigned		*sasl_maxbuf;
	Sockbuf_Buf		sec_buf_in;
	Sockbuf_Buf		buf_in;
	Sockbuf_Buf		buf_out;
};

static int
sb_sasl_setup( Sockbuf_IO_Desc *sbiod, void *arg )
{
	struct sb_sasl_data	*p;

	assert( sbiod != NULL );

	p = NSLBERI_MALLOC( sizeof( *p ) );
	if ( p == NULL ) {
		return -1;
	}
	p->sasl_context = (sasl_conn_t *)arg;
	ber_pvt_sb_buf_init( &p->sec_buf_in );
	ber_pvt_sb_buf_init( &p->buf_in );
	ber_pvt_sb_buf_init( &p->buf_out );
	if ( ber_pvt_sb_grow_buffer( &p->sec_buf_in, SASL_MIN_BUFF_SIZE ) < 0 ) {
		NSLBERI_FREE( p );
		errno = ENOMEM;
		return -1;
	}
	sasl_getprop( p->sasl_context, SASL_MAXOUTBUF,
		(SASL_CONST void **) &p->sasl_maxbuf );
	    
	sbiod->sbiod_pvt = p;

	return 0;
}

static int
sb_sasl_remove( Sockbuf_IO_Desc *sbiod )
{
	struct sb_sasl_data	*p;

	assert( sbiod != NULL );
	
	p = (struct sb_sasl_data *)sbiod->sbiod_pvt;
	/*
	 * SASLv2 encode/decode buffers are managed by
	 * libsasl2. Ensure they are not freed by liblber.
	 */
	p->buf_in.buf_base = NULL;
	p->buf_out.buf_base = NULL;
	ber_pvt_sb_buf_destroy( &p->sec_buf_in );
	ber_pvt_sb_buf_destroy( &p->buf_in );
	ber_pvt_sb_buf_destroy( &p->buf_out );
	NSLBERI_FREE( p );
	sbiod->sbiod_pvt = NULL;
	return 0;
}

static int
sb_sasl_pkt_length( const unsigned char *buf, int debuglevel )
{
	int		size;

	assert( buf != NULL );

	size = buf[0] << 24
		| buf[1] << 16
		| buf[2] << 8
		| buf[3];
   
	if ( size > SASL_MAX_BUFF_SIZE ) {
		size = 16; /* this should lead to an error. */
	}

	return size + 4; /* include the size !!! */
}

/* Drop a processed packet from the input buffer */
static void
sb_sasl_drop_packet ( Sockbuf_Buf *sec_buf_in, int debuglevel )
{
	int	len;

	len = sec_buf_in->buf_ptr - sec_buf_in->buf_end;
	if ( len > 0 ) {
		memcpy( sec_buf_in->buf_base, sec_buf_in->buf_base +
			sec_buf_in->buf_end, len );
	}
   
	if ( len >= 4 ) {
		sec_buf_in->buf_end = sb_sasl_pkt_length(
			(unsigned char *) sec_buf_in->buf_base, debuglevel);
	}
	else {
		sec_buf_in->buf_end = 0;
	}
	sec_buf_in->buf_ptr = len;
}

static int
sb_sasl_read( Sockbuf_IO_Desc *sbiod, void *buf, int len)
{
	struct sb_sasl_data	*p;
	int		ret, bufptr;
   
	assert( sbiod != NULL );
	assert( SOCKBUF_VALID( sbiod->sbiod_sb ) );

	p = (struct sb_sasl_data *)sbiod->sbiod_pvt;

	/* Are there anything left in the buffer? */
	ret = ber_pvt_sb_copy_out( &p->buf_in, buf, len );
	bufptr = ret;
	len -= ret;

	if ( len == 0 ) {
		return bufptr;
	}

	ber_pvt_sb_buf_init( &p->buf_in );

	/* Read the length of the packet */
	while ( p->sec_buf_in.buf_ptr < 4 ) {
		ret = LBER_SBIOD_READ_NEXT( sbiod, p->sec_buf_in.buf_base,
			4 - p->sec_buf_in.buf_ptr );
		if ( ret <= 0 || errno == EINTR) {
			return bufptr ? bufptr : ret;
		}

		p->sec_buf_in.buf_ptr += ret;
	}

	/* The new packet always starts at p->sec_buf_in.buf_base */
	ret = sb_sasl_pkt_length( 
		(unsigned char *) p->sec_buf_in.buf_base, 0 );

	/* Grow the packet buffer if neccessary */
	if ( ( p->sec_buf_in.buf_size < (int) ret ) && 
		ber_pvt_sb_grow_buffer( &p->sec_buf_in, ret ) < 0 )
	{
		errno = ENOMEM;
		return -1;
	}
	p->sec_buf_in.buf_end = ret;

	/* Did we read the whole encrypted packet? */
	while ( p->sec_buf_in.buf_ptr < p->sec_buf_in.buf_end ) {
		/* No, we have got only a part of it */
		ret = p->sec_buf_in.buf_end - p->sec_buf_in.buf_ptr;

		ret = LBER_SBIOD_READ_NEXT( sbiod, p->sec_buf_in.buf_base +
			p->sec_buf_in.buf_ptr, ret );
		if ( ret <= 0 || errno == EINTR) {
			return bufptr ? bufptr : ret;
		}

		p->sec_buf_in.buf_ptr += ret;
   	}

	/* Decode the packet */
	ret = sasl_decode( p->sasl_context, p->sec_buf_in.buf_base,
		p->sec_buf_in.buf_end,
		(SASL_CONST char **)&p->buf_in.buf_base,
		(unsigned *)&p->buf_in.buf_end );

	/* Drop the packet from the input buffer */
	sb_sasl_drop_packet( &p->sec_buf_in, 0 );

	if ( ret != SASL_OK ) {
		errno = EIO;
		return -1;
	}
	
	p->buf_in.buf_size = p->buf_in.buf_end;

	bufptr += ber_pvt_sb_copy_out( &p->buf_in, (char*) buf + bufptr, len );

	return bufptr;
}

static int
sb_sasl_write( Sockbuf_IO_Desc *sbiod, void *buf, int len)
{
	struct sb_sasl_data	*p;
	int			ret;

	assert( sbiod != NULL );
	assert( SOCKBUF_VALID( sbiod->sbiod_sb ) );

	p = (struct sb_sasl_data *)sbiod->sbiod_pvt;

	/* Are there anything left in the buffer? */
	if ( p->buf_out.buf_ptr != p->buf_out.buf_end ) {
		ret = ber_pvt_sb_do_write( sbiod, &p->buf_out );
		if ( ret < 0 ) {
			return ret;
		}
		/* Still have something left?? */
		if ( p->buf_out.buf_ptr != p->buf_out.buf_end ) {
			errno = EAGAIN;
			return 0;
		}
	}

	/* now encode the next packet. */
	ber_pvt_sb_buf_init( &p->buf_out );
	if ( len > *p->sasl_maxbuf - 100 ) {
		len = *p->sasl_maxbuf - 100;	/* For safety margin */
	}
	ret = sasl_encode( p->sasl_context, buf, len,
		(SASL_CONST char **)&p->buf_out.buf_base,
		(unsigned *)&p->buf_out.buf_size );
	if ( ret != SASL_OK ) {
		return -1;
	}
	p->buf_out.buf_end = p->buf_out.buf_size;

	ret = ber_pvt_sb_do_write( sbiod, &p->buf_out );
	if ( ret <= 0 ) {
		return ret;
	}
	return len;
}

static int
sb_sasl_ctrl( Sockbuf_IO_Desc *sbiod, int opt, void *arg )
{
	struct sb_sasl_data	*p;

	p = (struct sb_sasl_data *)sbiod->sbiod_pvt;

	if ( opt == LBER_SB_OPT_DATA_READY ) {
		if ( p->buf_in.buf_ptr != p->buf_in.buf_end ) {
			return 1;
		}
	}
	
	return LBER_SBIOD_CTRL_NEXT( sbiod, opt, arg );
}

int ldap_pvt_sasl_install( Sockbuf *sb, void *ctx_arg )
{

	VV_ldap_pvt_sockbuf_io_sasl.sbi_setup = &sb_sasl_setup;
	VV_ldap_pvt_sockbuf_io_sasl.sbi_remove = &sb_sasl_remove;
	VV_ldap_pvt_sockbuf_io_sasl.sbi_ctrl = &sb_sasl_ctrl;
	VV_ldap_pvt_sockbuf_io_sasl.sbi_read = &sb_sasl_read;
	VV_ldap_pvt_sockbuf_io_sasl.sbi_write = &sb_sasl_write;
	VV_ldap_pvt_sockbuf_io_sasl.sbi_close = NULL;

	/* don't install the stuff unless security has been negotiated */

	if ( !ber_sockbuf_ctrl( sb, LBER_SB_OPT_HAS_IO,
			&VV_ldap_pvt_sockbuf_io_sasl ) )
	{
		ber_sockbuf_add_io( sb, &VV_ldap_pvt_sockbuf_io_sasl,
			LBER_SBIOD_LEVEL_APPLICATION, ctx_arg );
	}

	return LDAP_SUCCESS;
}

void ldap_pvt_sasl_remove( Sockbuf *sb )
{
	ber_sockbuf_remove_io( sb, &VV_ldap_pvt_sockbuf_io_sasl,
		LBER_SBIOD_LEVEL_APPLICATION );
}

static int
sasl_err2ldap( int saslerr )
{
	int rc;

	switch (saslerr) {
		case SASL_CONTINUE:
			rc = LDAP_MORE_RESULTS_TO_RETURN;
			break;
		case SASL_INTERACT:
			rc = LDAP_LOCAL_ERROR;
			break;
		case SASL_OK:
			rc = LDAP_SUCCESS;
			break;
		case SASL_FAIL:
			rc = LDAP_LOCAL_ERROR;
			break;
		case SASL_NOMEM:
			rc = LDAP_NO_MEMORY;
			break;
		case SASL_NOMECH:
			rc = LDAP_AUTH_UNKNOWN;
			break;
		case SASL_BADAUTH:
			rc = LDAP_AUTH_UNKNOWN;
			break;
		case SASL_NOAUTHZ:
			rc = LDAP_PARAM_ERROR;
			break;
		case SASL_TOOWEAK:
		case SASL_ENCRYPT:
			rc = LDAP_AUTH_UNKNOWN;
			break;
		default:
			rc = LDAP_LOCAL_ERROR;
			break;
	}

	assert( rc == LDAP_SUCCESS || ( rc > 0 ) );
	return rc;
}

int
ldap_int_sasl_open(
	LDAP *ld, 
	LDAPConn *lc,
	const char * host,
        unsigned flags )
{
	int rc;
	sasl_conn_t *ctx;

	assert( lc->lconn_sasl_authctx == NULL );

	if ( host == NULL ) {
		ld->ld_errno = LDAP_LOCAL_ERROR;
		return ld->ld_errno;
	}

	if ( ldap_int_sasl_init() ) {
		ld->ld_errno = LDAP_LOCAL_ERROR;
		return ld->ld_errno;
	}

	rc = sasl_client_new( "ldap", host, NULL, NULL,
		client_callbacks, flags, &ctx );

	if ( rc != SASL_OK ) {
		ld->ld_errno = sasl_err2ldap( rc );
		return ld->ld_errno;
	}

	lc->lconn_sasl_authctx = ctx;

	return LDAP_SUCCESS;
}

int ldap_int_sasl_close( LDAP *ld, LDAPConn *lc )
{
	sasl_conn_t *ctx = lc->lconn_sasl_authctx;

	if( ctx != NULL ) {
		sasl_dispose( &ctx );
		if ( lc->lconn_sasl_sockctx &&
			lc->lconn_sasl_authctx != lc->lconn_sasl_sockctx ) {
			ctx = lc->lconn_sasl_sockctx;
			sasl_dispose( &ctx );
		}
		lc->lconn_sasl_sockctx = NULL;
		lc->lconn_sasl_authctx = NULL;
	}

	return LDAP_SUCCESS;
}

int
ldap_int_sasl_bind(
	LDAP			*ld,
	const char		*dn,
	const char		*mechs,
	LDAPControl		**sctrls,
	LDAPControl		**cctrls,
	struct sec_micro          *timeout,
	unsigned		flags,
	void                    *defaults,
        const char              *spn_service,
        const char              *spn_realm)
{
	const char *mech = NULL;
	const char *pmech = NULL;
	int			saslrc, rc;
	sasl_ssf_t		*ssf = NULL, external_ssf;
	sasl_conn_t	*ctx, *oldctx = NULL;
	sasl_interact_t *prompts = NULL;
	unsigned credlen;
	struct berval ccred;
	LBER_SOCKET		sd;
	sasl_security_properties_t secprops;
#ifdef NOT_USED_BY_NETAPP
	char *data;
	void	*ssl;
#endif

	/* do a quick !LDAPv3 check... ldap_sasl_bind will do the rest. */
	if (ld->ld_version < LDAP_VERSION3) {
		ld->ld_errno = LDAP_NOT_SUPPORTED;
		return ld->ld_errno;
	}

	if ( ld->ld_sbp == ((Sockbuf *)NULL) ) {
		sd = -1;
	}
	else {
		sd = ld->ld_sbp->sb_sd;
	}

	if ( sd == -1 ) {
 		/* not connected yet */
 		int rc;

		rc = nsldapi_open_ldap_defconn( ld );
		if( rc < 0 ) {
			return ld->ld_errno;
		}

		if ( ld->ld_sbp == ((Sockbuf *)NULL) ) {
			sd = -1;
		}
		else {
			sd = ld->ld_sbp->sb_sd;
		}

		if( sd == -1 ) {
			ld->ld_errno = LDAP_LOCAL_ERROR;
			return ld->ld_errno;
		}
	}   

	oldctx = ld->ld_defconn->lconn_sasl_authctx;

	/* If we already have an authentication context, clear it out */
	if( oldctx ) {
		if ( oldctx != ld->ld_defconn->lconn_sasl_sockctx ) {
			sasl_dispose( &oldctx );
		}
		ld->ld_defconn->lconn_sasl_authctx = NULL;
	}

	{ 
	/*
	char *saslhost = nsldapi_host_connected_to( ld->ld_sbp );
	*/
	char *saslhost = nsldapi_strdup( ld->ld_defhost );
	if ( saslhost == NULL ) {
		return LDAP_NO_MEMORY;
	}
	rc = ldap_int_sasl_open( ld, ld->ld_defconn, saslhost, flags );
		NSLDAPI_FREE( saslhost );
	}

	if ( rc != LDAP_SUCCESS ) {
		return rc;
	}

	ctx = ld->ld_defconn->lconn_sasl_authctx;

#ifdef NOT_USED_BY_NETAPP
	/* Check for TLS */
	ssl = ldap_pvt_tls_sb_ctx( ld->ld_sb );
	if ( ssl ) {
		struct berval authid = BER_BVNULL;
		int fac;

		fac = ldap_pvt_tls_get_strength( ssl );
		/* failure is OK, we just can't use SASL EXTERNAL */
		(void) ldap_pvt_tls_get_my_dn( ssl, &authid, NULL, 0 );

		(void) ldap_int_sasl_external( ld, ld->ld_defconn, authid.bv_val, fac );
		LDAP_FREE( authid.bv_val );
	}

#if !defined(_WIN32)
	/* Check for local */
	if ( ldap_pvt_url_scheme2proto( ld->ld_defconn->lconn_server->lud_scheme ) == LDAP_PROTO_IPC ) {
		char authid[sizeof("uidNumber=4294967295+gidNumber=4294967295,"
			"cn=peercred,cn=external,cn=auth")];
		sprintf( authid, "uidNumber=%d+gidNumber=%d,"
			"cn=peercred,cn=external,cn=auth",
			(int) geteuid(), (int) getegid() );
		(void) ldap_int_sasl_external( ld, ld->ld_defconn, authid, LDAP_PVT_SASL_LOCAL_SSF );
	}
#endif

#endif	/* NOT_USED_BY_NETAPP */

	external_ssf = LDAP_PVT_SASL_LOCAL_SSF;

	sasl_setprop( ctx, SASL_SSF_EXTERNAL,
		&external_ssf );
	

	/* Init defaults... */
	memset(&secprops, 0L, sizeof(secprops));
	secprops.maxbufsize = SASL_MAX_BUFF_SIZE;
	secprops.max_ssf = INT_MAX;
	secprops.security_flags = 
		SASL_SEC_NOPLAINTEXT | SASL_SEC_NOANONYMOUS;

	/* (re)set security properties */
	sasl_setprop( ctx, SASL_SEC_PROPS,
		&secprops );

	ccred.bv_val = NULL;
	ccred.bv_len = 0;

	/*
	 * Save the SPN info provided by the caller (if any).
	 */
	ctx->spn_service = spn_service;
	ctx->spn_realm   = spn_realm;

	do {
		saslrc = sasl_client_start( ctx,
			mechs,
			&prompts,
			(SASL_CONST char **)&ccred.bv_val,
			&credlen,
			&mech );

		if( pmech == NULL && mech != NULL ) {
			pmech = mech;
		}

	} while ( saslrc == SASL_INTERACT );

	ccred.bv_len = credlen;

	if ( (saslrc != SASL_OK) && (saslrc != SASL_CONTINUE) ) {
		rc = ld->ld_errno = sasl_err2ldap( saslrc );
		ld->ld_error = nsldapi_strdup(sasl_errdetail( ctx ));
		goto done;
	}

	do {
		struct berval *scred;
		unsigned credlen;

		scred = NULL;

		rc = ldap_sasl_bind_st( ld, dn, mech, &ccred, timeout, sctrls, cctrls, &scred );

		if ( ccred.bv_val != NULL ) {
			ccred.bv_val = NULL;
		}

		if ( rc != LDAP_SUCCESS && rc != LDAP_SASL_BIND_IN_PROGRESS ) {
			if( scred && scred->bv_len ) {
				ber_bvfree( scred );
			}
			rc = ld->ld_errno;
			goto done;
		}

		if( rc == LDAP_SUCCESS && saslrc == SASL_OK ) {
			/* we're done, no need to step */
			if( scred && scred->bv_len ) {
				/* 
				 * But server provided us with data!, so we
				 * have an error
				 */
				ber_bvfree( scred );
				rc = ld->ld_errno = LDAP_LOCAL_ERROR;
				goto done;
			}

			ber_bvfree( scred );
			break;
		}

		do {
			saslrc = sasl_client_step( ctx,
				(scred == NULL) ? NULL : scred->bv_val,
				(scred == NULL) ? 0 : scred->bv_len,
				&prompts,
				(SASL_CONST char **)&ccred.bv_val,
				&credlen );

		} while ( saslrc == SASL_INTERACT );

		ccred.bv_len = credlen;
		ber_bvfree( scred );

		if ( (saslrc != SASL_OK) && (saslrc != SASL_CONTINUE) ) {
			ld->ld_errno = sasl_err2ldap( saslrc );
			ld->ld_error = nsldapi_strdup(sasl_errdetail(ctx));
			rc = ld->ld_errno;
			goto done;
		}
	} while ( rc == LDAP_SASL_BIND_IN_PROGRESS );

	if ( rc != LDAP_SUCCESS ) {
		goto done;
	}

	if ( saslrc != SASL_OK ) {
		ld->ld_error = nsldapi_strdup(sasl_errdetail( ctx ));
		rc = ld->ld_errno = sasl_err2ldap( saslrc );
		goto done;
	}

	saslrc = sasl_getprop( ctx, SASL_SSF, (SASL_CONST void **) &ssf );
	if( saslrc == SASL_OK ) {

		if( ssf && *ssf ) {
			if ( ld->ld_defconn->lconn_sasl_sockctx ) {
				oldctx = ld->ld_defconn->lconn_sasl_sockctx;
				sasl_dispose( &oldctx );
				ldap_pvt_sasl_remove( ld->ld_sbp );
			}
			ldap_pvt_sasl_install( ld->ld_conns->lconn_sb, ctx );
			ld->ld_defconn->lconn_sasl_sockctx = ctx;
		}
	}
	ld->ld_defconn->lconn_sasl_authctx = ctx;

done:
	return rc;
}

sk_mutex_t *ldap_pvt_sasl_mutex_new(void)
{
	return ((sk_mutex_t*)sk_kmem_zalloc(sizeof(sk_mutex_t), M_LDAP, SKMA_SLEEP));
}

void ldap_pvt_sasl_mutex_dispose(sk_mutex_t *mutex)
{
	if (mutex) {
		sk_kmem_free(mutex);
	}
	return;
}

int ldap_pvt_sasl_mutex_lock(sk_mutex_t *mutex)
{
	if (!sk_is_mutex_mine(mutex)) {
		sk_mutex_lock(mutex);
	}
	return SASL_OK;
}

int ldap_pvt_sasl_mutex_unlock(sk_mutex_t *mutex)
{
	sk_mutex_unlock(mutex);
	return SASL_OK;
}

