/*
 *
 * Copyright (c) 1992-2006 Network Appliance, Inc.
 * All rights reserved.
 */
/*
 * Abstract Syntax Notation One, ASN.1
 * As defined in ISO/IS 8824 and ISO/IS 8825
 * This implements a subset of the above International Standards that
 * is sufficient to implement SNMP.
 *
 * Encodes abstract data types into a machine independent stream of bytes.
 *
 */
/**********************************************************************
	Copyright 1988, 1989, 1991, 1992 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/
/**
 * @defgroup asn1_packet_parse asn1 parsing and datatype manipulation routines.
 * @ingroup library
 *
 * @{
 * 
 * Note on 
 * 
 * Re-allocating reverse ASN.1 encoder functions.  Synopsis:
 *
 * \code
 *
 * uint8_t *buf = (uint8_t*)malloc(100);
 * uint8_t type = (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER);
 * size_t buf_len = 100, offset = 0;
 * long data = 12345;
 * int allow_realloc = 1;
 * 
 * if (asn_realloc_rbuild_int(&buf, &buf_len, &offset, allow_realloc,
 *                            type, &data, sizeof(long)) == 0) {
 *     error;
 * }
 * 
 * \endcode
 *
 * NOTE WELL: after calling one of these functions with allow_realloc
 * non-zero, buf might have moved, buf_len might have grown and
 * offset will have increased by the size of the encoded data.
 * You should **NEVER** do something like this:
 * 
 * \code
 *
 * uint8_t *buf = (uint8_t *)malloc(100), *ptr;
 * uint8_t type = (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER);
 * size_t buf_len = 100, offset = 0;
 * long data1 = 1234, data2 = 5678;
 * int rc = 0, allow_realloc = 1;
 * 
 * rc  = asn_realloc_rbuild_int(&buf, &buf_len, &offset, allow_realloc,
 *                                type, &data1, sizeof(long));
 * ptr = buf[buf_len - offset];   / * points at encoding of data1 * /
 * if (rc == 0) {
 *      error;
 * }
 * rc  = asn_realloc_rbuild_int(&buf, &buf_len, &offset, allow_realloc,
 *                              type, &data2, sizeof(long));
 * make use of ptr here;
 * 
 * \endcode
 * 
 * ptr is **INVALID** at this point.  In general, you should store the
 * offset value and compute pointers when you need them:
 * 
 * 
 * \code
 *
 * uint8_t *buf = (uint8_t *)malloc(100), *ptr;
 * uint8_t type = (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER);
 * size_t buf_len = 100, offset = 0, ptr_offset;
 * long data1 = 1234, data2 = 5678;
 * int rc = 0, allow_realloc = 1;
 * 
 * rc  = asn_realloc_rbuild_int(&buf, &buf_len, &offset, allow_realloc,
 *                              type, &data1, sizeof(long));
 * ptr_offset = offset;
 * if (rc == 0) {
 *      error;
 * }
 * rc  = asn_realloc_rbuild_int(&buf, &buf_len, &offset, allow_realloc,
 *                              type, &data2, sizeof(long));
 * ptr = buf + buf_len - ptr_offset
 * make use of ptr here;
 * 
 * \endcode
 * 
 * 
 * Here, you can see that ptr will be a valid pointer even if the block of
 * memory has been moved, as it may well have been.  Plenty of examples of
 * usage all over asn1.c, snmp_api.c, snmpusm.c.
 * 
 * The other thing you should **NEVER** do is to pass a pointer to a buffer
 * on the stack as the first argument when allow_realloc is non-zero, unless
 * you really know what you are doing and your machine/compiler allows you to
 * free non-heap memory.  There are rumours that such things exist, but many
 * consider them no more than the wild tales of a fool.
 * 
 * Of course, you can pass allow_realloc as zero, to indicate that you do not
 * wish the packet buffer to be reallocated for some reason; perhaps because
 * it is on the stack.  This may be useful to emulate the functionality of
 * the old API:
 *
 * \code 
 * 
 * uint8_t my_static_buffer[100], *cp = NULL;
 * size_t my_static_buffer_len = 100;
 * float my_pi = (float)22/(float)7;
 * 
 * cp = asn_rbuild_float(my_static_buffer, &my_static_buffer_len,
 *                       ASN_OPAQUE_FLOAT, &my_pi, sizeof(float));
 * if (cp == NULL) {
 * error;
 * }
 * 
 * \endcode
 * 
 * IS EQUIVALENT TO:
 * 
 * \code
 * 
 * uint8_t my_static_buffer[100];
 * size_t my_static_buffer_len = 100, my_offset = 0;
 * float my_pi = (float)22/(float)7;
 * int rc = 0;
 * 
 * rc = asn_realloc_rbuild_float(&my_static_buffer, &my_static_buffer_len,
 *                               &my_offset, 0,
 *                               ASN_OPAQUE_FLOAT, &my_pi, sizeof(float));
 * if (rc == 0) {
 *   error;
 * }
 * \endcode
 * 
 */


#include <net-snmp/net-snmp-config.h>


#ifdef KINETICS
#include "gw.h"
#endif

#ifdef HAVE_STRING_H
#include <legacy/string.h>
#else
#include <strings.h>
#endif

#include <sys/types.h>
#ifdef HAVE_STDLIB_H
#include <legacy/stdlib.h>
#endif
#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <net7g/netinet/sk_in.h>
#endif

#ifdef vms
#include <in.h>
#endif

#ifdef HAVE_DMALLOC_H
#include <dmalloc.h>
#endif

#include <net-snmp/output_api.h>
#include <net-snmp/utilities.h>

#include <net-snmp/library/asn1.h>
#include <net-snmp/library/int64.h>
#include <net-snmp/library/mib.h>

#include <sys/_null.h>

#include <net-snmp/library/snmp_api.h>

#include <sys/param.h>
#include <sys/cdefs.h>
#include <net7g/sys/sk_socket.h>
#include <net7g/sys/sk_mbuf.h>
#include <sk/sk_domain.h>
#include <sk/sk_common.h>
#include <ontap/ems/snmp_parse_ems.h>

/**
 * @internal
 * output an error for a wrong size
 * 
 * @param str        error string
 * @param wrongsize  wrong size
 * @param rightsize  expected size
 */
static
    void
_asn_size_err(const char *str, size_t wrongsize, size_t rightsize)
{
    char            ebuf[128];

    snprintf(ebuf, sizeof(ebuf),
            "%s size %lu: s/b %lu", str,
	    (unsigned long)wrongsize, (unsigned long)rightsize);
    ebuf[ sizeof(ebuf)-1 ] = 0;
    EMS_ems_snmp_parse_error(ebuf);
}

/**
 * @internal 
 * output an error for a wrong length
 * 
 * @param str        error string
 * @param wrongsize  wrong  length
 * @param rightsize  expected length
 */
static
    void
_asn_length_err(const char *str, size_t wrongsize, size_t rightsize)
{
    char            ebuf[128];

    snprintf(ebuf, sizeof(ebuf),
            "%s length %lu too large: exceeds %lu", str,
	    (unsigned long)wrongsize, (unsigned long)rightsize);
    ebuf[ sizeof(ebuf)-1 ] = 0;
    EMS_ems_snmp_parse_error(ebuf);
}

/**
 * @internal
 * call after asn_parse_length to verify result.
 * 
 * @param str  error string
 * @param bufp start of buffer
 * @param data start of data
 * @param plen  ?
 * @param dlen  ?
 * 
 * @return 1 on error 0 on success
 */
static
    int
_asn_parse_length_check(const char *str,
                        uint8_t * bufp, uint8_t * data,
                        uint32_t plen, size_t dlen)
{
    char            ebuf[128];
    size_t          header_len;

    if (bufp == NULL) {
        /*
         * error message is set 
         */
        return 1;
    }
    header_len = bufp - data;
    if (plen > 0x7fffffff || header_len > 0x7fffffff ||
        ((size_t) plen + header_len) > dlen) {
        snprintf(ebuf, sizeof(ebuf),
                "%s: message overflow: %d len + %d delta > %d len",
                str, (int) plen, (int) header_len, (int) dlen);
        ebuf[ sizeof(ebuf)-1 ] = 0;
        EMS_ems_snmp_parse_error(ebuf);
        return 1;
    }
    return 0;
}


/**
 * @internal 
 * call after asn_build_header to verify result.
 * 
 * @param str     error string to output
 * @param data    data pointer to verify (NULL => error )
 * @param datalen  data len to check
 * @param typedlen  type length
 * 
 * @return 0 on success, 1 on error
 */
static
    int
_asn_build_header_check(const char *str, uint8_t * data,
                        size_t datalen, size_t typedlen)
{
    char            ebuf[128];

    if (data == NULL) {
        /*
         * error message is set 
         */
        return 1;
    }
    if (datalen < typedlen) {
        snprintf(ebuf, sizeof(ebuf),
                "%s: bad header, length too short: %lu < %lu", str,
                (unsigned long)datalen, (unsigned long)typedlen);
        ebuf[ sizeof(ebuf)-1 ] = 0;
        EMS_ems_snmp_parse_error(ebuf);
        return 1;
    }
    return 0;
}

/**
 * @internal 
 * checks the incoming packet for validity and returns its size or 0 
 * 
 * @param pkt The packet 
 * @param len The length to check 
 * 
 * @return The size of the packet if valid; 0 otherwise
 */
int
asn_check_packet(uint8_t * pkt, size_t len)
{
    uint32_t          asn_length;

    if (len < 2)
        return 0;               /* always too short */

    if (*pkt != (uint8_t) (ASN_SEQUENCE | ASN_CONSTRUCTOR))
        return -1;              /* wrong type */

    if (*(pkt + 1) & 0x80) {
        /*
         * long length 
         */
        if ((int) len < (int) (*(pkt + 1) & ~0x80) + 2)
            return 0;           /* still to short, incomplete length */
        asn_parse_length(pkt + 1, &asn_length);
        return (asn_length + 2 + (*(pkt + 1) & ~0x80));
    } else {
        /*
         * short length 
         */
        return (*(pkt + 1) + 2);
    }
}

static
    int
_asn_bitstring_check(const char *str, size_t asn_length, uint8_t datum)
{
    char            ebuf[128];

    if (asn_length < 1) {
        snprintf(ebuf, sizeof(ebuf),
                "%s: length %d too small", str, (int) asn_length);
        ebuf[ sizeof(ebuf)-1 ] = 0;
        EMS_ems_snmp_parse_error(ebuf);
        return 1;
    }
    /*
     * if (datum > 7){
     * sprintf(ebuf,"%s: datum %d >7: too large", str, (int)(datum));
     * ERROR_MSG(ebuf);
     * return 1;
     * }
     */
    return 0;
}

/**
 * @internal 
 * asn_parse_int - pulls a long out of an int type.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */

uint8_t *
asn_parse_int(
	uint8_t         *data,	/* IN - pointer to start of object */
	size_t         *datalength,	/* IN/OUT - number of valid bytes
					 * left in buffer */
	uint8_t         *type,	/* OUT - asn type of object */
	int32_t           *intp,	/* IN/OUT - pointer to start of output buffer */
	size_t          intsize/* IN - size of output buffer */
)
{
	/*
	 * ASN.1 integer ::= 0x02 asnlength byte {byte}*
	 */
	static const char *errpre = "parse int";
	register uint8_t *bufp = data;
	uint32_t          asn_length;
	register int32_t   value = 0;

	if (intsize != sizeof(int32_t)) {
		_asn_size_err(errpre, intsize, sizeof(int32_t));
		return NULL;
	}
	*type = *bufp++;
	bufp = asn_parse_length(bufp, &asn_length);
	if (_asn_parse_length_check
	    (errpre, bufp, data, asn_length, *datalength)) {
		return NULL;
	}

	if ((size_t)asn_length > intsize) {
		_asn_length_err(errpre, (size_t) asn_length, intsize);
		return NULL;
	}

	*datalength -= (int) asn_length + (bufp - data);
	if (*bufp & 0x80)
		value = -1;	/* integer is negative */

	DEBUGDUMPSETUP("recv", data, bufp - data + asn_length);

	while (asn_length--)
		value = (value << 8) | *bufp++;

	DEBUGMSG(("dumpv_recv", "  Integer:\t%ld (0x%.2X)\n", value, value));

	*intp = value;
	return bufp;
}


/**
 * @internal 
 * asn_parse_unsigned_int - pulls an uint32_t out of an ASN int type.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
uint8_t *
asn_parse_unsigned_int(
	uint8_t         *data,	/* IN - pointer to start of object */
	size_t         *datalength,	/* IN/OUT - number of valid bytes
					 * left in buffer */
	uint8_t         *type,	/* OUT - asn type of object */
	uint32_t          *intp,	/* IN/OUT - pointer to start of output buffer */
	size_t          intsize/* IN - size of output buffer */
)
{
	/*
	 * ASN.1 integer ::= 0x02 asnlength byte {byte}*
	 */
	static const char *errpre = "parse uint32_t";
	register uint8_t *bufp = data;
	uint32_t          asn_length;
	register uint32_t value = 0;

	if (intsize != sizeof(int32_t)) {
		_asn_size_err(errpre, intsize, sizeof(int32_t));
		return NULL;
	}
	*type = *bufp++;
	bufp = asn_parse_length(bufp, &asn_length);
	if (_asn_parse_length_check
	    (errpre, bufp, data, asn_length, *datalength)) {
		return NULL;
	}
	if ((asn_length > (intsize + 1)) ||
	    ((asn_length == intsize + 1) && *bufp != 0x00)) {
		_asn_length_err(errpre, (size_t) asn_length, intsize);
		return NULL;
	}
	*datalength -= (int) asn_length + (bufp - data);
	if (*bufp & 0x80)
		value = (uint32_t) -1;	/* integer is negative */

	DEBUGDUMPSETUP("recv", data, bufp - data + asn_length);

	while (asn_length--)
		value = (value << 8) | *bufp++;

	DEBUGMSG(("dumpv_recv", "  UInteger:\t%ld (0x%.2X)\n", value, value));

	*intp = value;
	return bufp;
}


/**
 * @internal 
 * asn_build_int - builds an ASN object containing an integer.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
SK_DOMAIN_EXPORT(asn_build_int, SK_REASON_MPSAFE);

uint8_t *
asn_build_int(
	uint8_t         *data,	/* IN - pointer to start of output buffer */
	size_t         *datalength,	/* IN/OUT - number of valid bytes
					 * left in buffer */
	uint8_t          type,	/* IN - asn type of object */
	int32_t           *intp,	/* IN - pointer to start of long integer */
	size_t          intsize/* IN - size of *intp */
)
{
	/*
	 * ASN.1 integer ::= 0x02 asnlength byte {byte}*
	 */

	static const char *errpre = "build int";
	register int32_t   integer;
	register uint32_t mask;
#ifndef SNMP_NO_DEBUGGING
	uint8_t         *initdatap = data;
#endif

	if (intsize != sizeof(int32_t)) {
		_asn_size_err(errpre, intsize, sizeof(int32_t));
		return NULL;
	}
	integer = *intp;
	/*
	 * Truncate "unnecessary" bytes off of the most significant end of
	 * this 2's complement integer. There should be no sequence of 9
	 * consecutive 1's or 0's at the most significant end of the integer.
	 */
	mask = 0x1FF << ((8 * (sizeof(int32_t) - 1)) - 1);
	/* mask is 0xFF800000 on a big-endian machine */
	while ((((integer & mask) == 0) || ((integer & mask) == mask)) && intsize > 1) {
		intsize--;
		integer <<= 8;
	}
	data = asn_build_header(data, datalength, type, intsize);
	if (_asn_build_header_check(errpre, data, *datalength, intsize)) {
		return NULL;
	}
	*datalength -= intsize;
	mask = 0xFF << (8 * (sizeof(int32_t) - 1));
	/* mask is 0xFF000000 on a big-endian machine */
	while (intsize--) {
		*data++ = (uint8_t) ((integer & mask) >> (8 * (sizeof(int32_t) - 1)));
		integer <<= 8;
	}
	DEBUGDUMPSETUP("send", initdatap, data - initdatap);
	DEBUGMSG(("dumpv_send", "  Integer:\t%ld (0x%.2X)\n", *intp, *intp));
	return data;
}

/**
 * @internal 
 * asn_build_unsigned_int - builds an ASN object containing an integer.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
SK_DOMAIN_EXPORT(asn_build_unsigned_int, SK_REASON_MPSAFE);
uint8_t *
asn_build_unsigned_int(
	uint8_t         *data,	/* IN - pointer to start of output buffer */
	size_t         *datalength,	/* IN/OUT - number of valid bytes
					 * left in buffer */
	uint8_t          type,	/* IN - asn type of object */
	uint32_t          *intp,	/* IN - pointer to start of long integer */
	size_t             intsize/* IN - size of *intp */
)
{
	/*
	 * ASN.1 integer ::= 0x02 asnlength byte {byte}*
	 */

	static const char *errpre = "build uint32_t";
	register uint32_t integer;
	register uint32_t mask;
	int             add_null_byte = 0;
#ifndef SNMP_NO_DEBUGGING
	uint8_t         *initdatap = data;
#endif

	if (intsize != sizeof(int32_t)) {
		_asn_size_err(errpre, intsize, sizeof(int32_t));
		return NULL;
	}
	integer = *intp;

	mask = 0xFF << (8 * (sizeof(int32_t) - 1));
	/*
	 * mask is 0xFF000000 on a big-endian machine 
	 */
	if ((uint8_t) ((integer & mask) >> (8 * (sizeof(int32_t) - 1))) & 0x80) {
		/*
		 * if MSB is set 
		 */
		add_null_byte = 1;
		intsize++;
	} else {
	    /*
	     * Truncate "unnecessary" bytes off of the most significant end of
	     * this 2's complement integer. There should be no sequence of 9
	     * consecutive 1's or 0's at the most significant end of the 
	     * integer.
	     */
		mask = 0x1FF << ((8 * (sizeof(int32_t) - 1)) - 1);
		/*
		 * mask is 0xFF800000 on a big-endian machine 
		 */
		while ((((integer & mask) == 0) || ((integer & mask) == mask))
		       && intsize > 1) {
			intsize--;
			integer <<= 8;
		}
	}
	data = asn_build_header(data, datalength, type, intsize);
	if (_asn_build_header_check(errpre, data, *datalength, intsize)) {
		return NULL;
	}

	*datalength -= intsize;
	if (add_null_byte == 1) {
		*data++ = '\0';
		intsize--;
	}
	mask = 0xFF << (8 * (sizeof(int32_t) - 1));
	/*
	 * mask is 0xFF000000 on a big-endian machine 
	 */
	while (intsize--) {
		*data++ = (uint8_t) ((integer & mask) >> (8 * (sizeof(int32_t) - 1)));
		integer <<= 8;
	}
	DEBUGDUMPSETUP("send", initdatap, data - initdatap);
	DEBUGMSG(("dumpv_send", "  UInteger:\t%ld (0x%.2X)\n", *intp, *intp));
	return data;
}


/**
 * @internal 
 * asn_parse_string - pulls an octet string out of an ASN octet string type.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  "string" is filled with the octet string.
 * ASN.1 octet string   ::=      primstring | cmpdstring
 * primstring           ::= 0x04 asnlength byte {byte}*
 * cmpdstring           ::= 0x24 asnlength string {string}*
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
uint8_t *
asn_parse_string(
	uint8_t         *data,	/* IN - pointer to start of object */
	size_t         *datalength,	/* IN/OUT - number of valid bytes
					 * left in buffer */
	uint8_t         *type,	/* OUT - asn type of object */
	uint8_t         *str,	/* IN/OUT - pointer to start of output buffer */
	size_t         *strlength	/* IN/OUT - size of output buffer */
)
{
	/*
	 * ASN.1 octet string ::= primstring | cmpdstring primstring ::= 0x04
	 * asnlength byte {byte}* cmpdstring ::= 0x24 asnlength string
	 * {string}* This doesn't yet support the compound string.
	 */
	static const char *errpre = "parse string";
	register uint8_t *bufp = data;
	uint32_t          asn_length;

	*type = *bufp++;
	bufp = asn_parse_length(bufp, &asn_length);
	if (_asn_parse_length_check
	    (errpre, bufp, data, asn_length, *datalength)) {
		return NULL;
	}
	if (asn_length > *strlength) {
		_asn_length_err(errpre, (size_t) asn_length, *strlength);
		return NULL;
	}

	DEBUGDUMPSETUP("recv", data, bufp - data + asn_length);

	bcopy((char *) bufp, (char *) str, (int) asn_length);
	*strlength = (int) asn_length;
	*datalength -= (int) asn_length + (bufp - data);

	DEBUGIF("dumpv_recv") {
		uint8_t         *buf = (uint8_t *) malloc(1 + asn_length);
		size_t          l = (buf != NULL) ? (1 + asn_length) : 0, ol = 0;
		
		if (sprint_realloc_asciistring
		    (&buf, &l, &ol, 1, str, asn_length)) {
			DEBUGMSG(("dumpv_recv", "  String:\t%s\n", buf));
		} else {
			if (buf == NULL) {
				DEBUGMSG(("dumpv_recv", "  String:\t[TRUNCATED]\n"));
			} else {
				DEBUGMSG(("dumpv_recv", "  String:\t%s [TRUNCATED]\n",
					  buf));
			}
		}
		if (buf != NULL) {
			free(buf);
		}
	}

	return bufp + asn_length;
}

/**
 * @internal
 * asn_build_string - Builds an ASN octet string object containing the input string.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */

SK_DOMAIN_EXPORT(asn_build_string, SK_REASON_MPSAFE);

uint8_t *
asn_build_string(
	uint8_t         *data,	/* IN - pointer to start of object */
	size_t         *datalength,	/* IN/OUT - number of valid bytes
					 * left in buffer */
	uint8_t          type,	/* IN - ASN type of string */
	uint8_t         *str,	/* IN - pointer to start of input buffer */
	size_t          strlength	/* IN - size of input buffer */
)
{
	/*
	 * ASN.1 octet string ::= primstring | cmpdstring
	 * primstring ::= 0x04 asnlength byte {byte}*
	 * cmpdstring ::= 0x24 asnlength string {string}*
	 * This code will never send a compound string.
	 */
#ifndef SNMP_NO_DEBUGGING
	uint8_t         *initdatap = data;
#endif
	data = asn_build_header(data, datalength, type, strlength);
	if (_asn_build_header_check
	    ("build string", data, *datalength, strlength)) {
		return NULL;
	}
	bcopy((char *) str, (char *) data, strlength);
	*datalength -= strlength;

	DEBUGDUMPSETUP("send", initdatap, data - initdatap + strlength);
	DEBUGIF("dumpv_send") {
		uint8_t         *buf = (uint8_t *) malloc(1 + strlength);
		size_t          l = (buf != NULL) ? (1 + strlength) : 0, ol = 0;
		
		if (sprint_realloc_asciistring
		    (&buf, &l, &ol, 1, str, strlength)) {
			DEBUGMSG(("dumpv_send", "  String:\t%s\n", buf));
		} else {
			if (buf == NULL) {
				DEBUGMSG(("dumpv_send", "  String:\t[TRUNCATED]\n"));
			} else {
				DEBUGMSG(("dumpv_send", "  String:\t%s [TRUNCATED]\n",
					  buf));
			}
		}
		if (buf != NULL) {
			free(buf);
		}
	}
	return data + strlength;
}

/**
 * @internal
 * asn_parse_header - interprets the ID and length of the current object.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   in this object following the id and length.
 *
 * @return  Returns a pointer to the first byte of the contents of this object.
 *          Returns NULL on any error.
 *
 */
uint8_t *
asn_parse_header(
	uint8_t         *data,	/* IN - pointer to start of object */
	size_t         *datalength,	/* IN/OUT - number of valid bytes
					 * left in buffer */
	uint8_t         *type	/* OUT - ASN type of object */
)
{
	register uint8_t *bufp = data;
	uint32_t          asn_length;

	if (!data || !datalength || !type) {
		ERROR("parse header: NULL pointer");
		return NULL;
	}
	bufp = data;
	/*
	 * this only works on data types < 30, i.e. no extension octets 
	 */
	if (IS_EXTENSION_ID(*bufp)) {
		ERROR("can't process ID >= 30");
		return NULL;
	}
	*type = *bufp;
	bufp = asn_parse_length(bufp + 1, &asn_length);
	if (_asn_parse_length_check
	    ("parse header", bufp, data, asn_length, *datalength)) {
		return NULL;
	}

	DEBUGDUMPSETUP("recv", data, (bufp - data));
	DEBUGMSG(("dumpv_recv", "  Header: 0x%.2X, len = %d (0x%X)\n", *data,
		  asn_length, asn_length));

	*datalength = (int) asn_length;
	return bufp;
}

/**
 * @internal
 * same as asn_parse_header with test for expected type
 *
 * @see asn_parse_header
 *
 * @param data          IN - pointer to start of object
 * @param datalength    IN/OUT - number of valid bytes left in buffer
 * @param type          OUT - asn type of object
 * @param expected_type IN expected type
 * @return  Returns a pointer to the first byte of the contents of this object.
 *          Returns NULL on any error.
 *
 */
uint8_t         *
asn_parse_sequence(uint8_t * data, size_t * datalength, uint8_t * type, 
		   uint8_t expected_type,     /* must be this type */
                   const char *estr)
{                               /* error message prefix */
    data = asn_parse_header(data, datalength, type);
    if (data && (*type != expected_type)) {
	    char            ebuf[128];
	    snprintf(ebuf, sizeof(ebuf),
		     "%s header type %02X: s/b %02X", estr,
		     (uint8_t) * type, (uint8_t) expected_type);
	    ebuf[ sizeof(ebuf)-1 ] = 0;
	    ERROR(ebuf);
	    return NULL;
    }
    return data;
}

/**
 * @internal
 * asn_build_header - builds an ASN header for an object with the ID and
 * length specified.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   in this object following the id and length.
 *
 *  This only works on data types < 30, i.e. no extension octets.
 *  The maximum length is 0xFFFF;
 *
 *  Returns a pointer to the first byte of the contents of this object.
 *  Returns NULL on any error.
 */
SK_DOMAIN_EXPORT(asn_build_header, SK_REASON_MPSAFE);

uint8_t *
asn_build_header(
	uint8_t         *data,	/* IN - pointer to start of object */
	size_t         *datalength,	/* IN/OUT - number of valid bytes
					 * left in buffer */
	uint8_t          type,	/* IN - ASN type of object */
	size_t          length	/* IN - length of object */
)
{
	char            ebuf[128];

	if (*datalength < 1) {
		snprintf(ebuf, sizeof(ebuf),
			 "bad header length < 1 :%lu, %lu",
			 (unsigned long)*datalength, (unsigned long)length);
		ebuf[ sizeof(ebuf)-1 ] = 0;
		ERROR(ebuf);
		return NULL;
	}
	*data++ = type;
	(*datalength)--;
	return asn_build_length(data, datalength, length);

}

/**
 * @internal
 * asn_build_sequence - builds an ASN header for a sequence with the ID and
 * length specified.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   in this object following the id and length.
 *
 *  This only works on data types < 30, i.e. no extension octets.
 *  The maximum length is 0xFFFF;
 *
 *  Returns a pointer to the first byte of the contents of this object.
 *  Returns NULL on any error.
 *
 * @param data         IN - pointer to start of object
 * @param datalength   IN/OUT - number of valid bytes left in buffer
 * @param type         IN - asn type of object
 * @param length       IN - length of object
 *
 * @return Returns a pointer to the first byte of the contents of this object.
 *         Returns NULL on any error.
 */
uint8_t         *
asn_build_sequence(uint8_t * data,
                   size_t * datalength, uint8_t type, size_t length)
{
	static const char *errpre = "build seq";
	char            ebuf[128];
	
	if (*datalength < 4) {
		snprintf(ebuf, sizeof(ebuf),
			 "%s: length %d < 4: PUNT", errpre,
			 (int) *datalength);
		ebuf[ sizeof(ebuf)-1 ] = 0;
		ERROR(ebuf);
		return NULL;
	}
	*datalength -= 4;
	*data++ = type;
	*data++ = (uint8_t) (0x02 | ASN_LONG_LEN);
	*data++ = (uint8_t) ((length >> 8) & 0xFF);
	*data++ = (uint8_t) (length & 0xFF);
	return data;
}

/*
 * asn_change_header_type - modifies an existing ASN header to change
 * the ASN type of the object; every other field is left unchanged.
 * On entry, datalength is input as the number of valid bytes following
 * "data".
 * 
 * This only works on data types < 30, i.e. no extension octets. The maximum
 * length is 0xFFFF.
 *
 * Returns TRUE on success, FALSE on error.
 */
int
asn_change_header_type(
	uint8_t         *data,	/* IN - pointer to start of object */
	size_t          datalength,	/* IN - number of valid bytes
					 * left in buffer */
	uint8_t          type	/* IN - ASN type of object */
)
{
	if (datalength < 1)
		return false;
	*data = type;
	return true;
}

/**
 * @internal
 * asn_parse_length - interprets the length of the current object.
 *
 *  On exit, length contains the value of this length field.
 *
 *  Returns a pointer to the first byte after this length
 *  field (aka: the start of the data field).
 *  Returns NULL on any error.
 */
uint8_t *
asn_parse_length(
	uint8_t         *data,	/* IN - pointer to start of length field */
	uint32_t          *length	/* OUT - value of length field */
)
{
	static const char *errpre = "parse length";
	char            ebuf[128];
	register uint8_t lengthbyte;

	if (!data || !length) {
		ERROR("parse length: NULL pointer");
		return NULL;
	}

	lengthbyte = *data;
	if (lengthbyte & ASN_LONG_LEN) {
		lengthbyte &= ~ASN_LONG_LEN;	/* turn MSb off */
		if (lengthbyte == 0) {
			snprintf(ebuf, sizeof(ebuf),
				 "%s: indefinite length not supported", errpre);
			ebuf[ sizeof(ebuf)-1 ] = 0;
			ERROR(ebuf);
			return NULL;
		}
		if (lengthbyte > sizeof(int32_t)) {
			snprintf(ebuf, sizeof(ebuf),
				 "%s: data length %d > %lu not supported", errpre,
				 lengthbyte, (unsigned long)sizeof(int32_t));
			ebuf[ sizeof(ebuf)-1 ] = 0;
			ERROR(ebuf);
			return NULL;
		}
		data++;
		*length = 0;            /* protect against short lengths */
		while (lengthbyte--) {
			*length <<= 8;
			*length |= *data++;
		}
		if ((int32_t) *length < 0) {
			snprintf(ebuf, sizeof(ebuf),
				 "%s: negative data length %ld\n", errpre,
				 (long) *length);
			ebuf[ sizeof(ebuf)-1 ] = 0;
                        EMS_ems_snmp_parse_error(ebuf);
			return NULL;
		}
		return data;
	} else {		/* short asnlength */
		*length = (int32_t) lengthbyte;
		return data + 1;
	}
}

/**
 * @internal
 * asn_build_length - builds an ASN header for a length with
 * length specified.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   in this object following the length.
 *
 *  Returns a pointer to the first byte of the contents of this object.
 *  Returns NULL on any error.
 */
SK_DOMAIN_EXPORT(asn_build_length, SK_REASON_MPSAFE);
uint8_t *
asn_build_length(
	uint8_t         *data,	/* IN - pointer to start of object */
	size_t         *datalength,	/* IN/OUT - number of valid bytes
					 * left in buffer */
	size_t          length	/* IN - length of object */
)
{
	static const char *errpre = "build length";
	char            ebuf[128];
	uint8_t         *start_data = data;

	/*
	 * no indefinite lengths sent 
	 */
	if (length < 0x80) {
		if (*datalength < 1) {
			snprintf(ebuf, sizeof(ebuf),
				 "%s: bad length < 1 :%lu, %lu", errpre,
				 (unsigned long)*datalength, (unsigned long)length);
			ebuf[ sizeof(ebuf)-1 ] = 0;
			ERROR(ebuf);
			return NULL;
		}
		*data++ = (uint8_t) length;
	} else if (length <= 0xFF) {
		if (*datalength < 2) {
			snprintf(ebuf, sizeof(ebuf),
				 "%s: bad length < 2 :%lu, %lu", errpre,
				 (unsigned long)*datalength, (unsigned long)length);
			ebuf[ sizeof(ebuf)-1 ] = 0;
			ERROR(ebuf);
			return NULL;
		}
		*data++ = (uint8_t) (0x01 | ASN_LONG_LEN);
		*data++ = (uint8_t) length;
	} else {		/* 0xFF < length <= 0xFFFF */
		if (*datalength < 3) {
			snprintf(ebuf, sizeof(ebuf),
				 "%s: bad length < 3 :%lu, %lu", errpre,
				 (unsigned long)*datalength, (unsigned long)length);
			ebuf[ sizeof(ebuf)-1 ] = 0;
			ERROR(ebuf);
			return NULL;
		}
		*data++ = (uint8_t) (0x02 | ASN_LONG_LEN);
		*data++ = (uint8_t) ((length >> 8) & 0xFF);
		*data++ = (uint8_t) (length & 0xFF);
	}
	*datalength -= (data - start_data);
	return data;

}

/**
 * @internal
 * asn_parse_objid - pulls an object indentifier out of an ASN object identifier type.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  "objid" is filled with the object identifier.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
uint8_t *
asn_parse_objid(
	uint8_t         *data,	/* IN - pointer to start of object */
	size_t         *datalength,	/* IN/OUT - number of valid bytes
					 * left in buffer */
	uint8_t         *type,	/* OUT - ASN type of object */
	oid            *objid,	/* IN/OUT - pointer to start of output buffer */
	size_t         *objidlength	/* IN/OUT - number of sub-id's in
					 * objid */
)
{
	/*
	 * ASN.1 objid ::= 0x06 asnlength subidentifier {subidentifier}*
	 * subidentifier ::= {leadingbyte}* lastbyte
	 * leadingbyte ::= 1 7bitvalue
	 * lastbyte ::= 0 7bitvalue
	 */
	register uint8_t *bufp = data;
	register oid   *oidp = objid + 1;
	register uint32_t  subidentifier;
	register int32_t   length;
	uint32_t           asn_length;

	*type = *bufp++;
	bufp = asn_parse_length(bufp, &asn_length);
	if (_asn_parse_length_check("parse objid", bufp, data,
				    asn_length, *datalength)) {
		return NULL;
	}
	*datalength -= (int) asn_length + (bufp - data);

	DEBUGDUMPSETUP("recv", data, bufp - data + asn_length);

	/*
	 * Handle invalid oid encodings of the form 06 00 robustly 
	 */
	if (asn_length == 0) {
		objid[0] = objid[1] = 0;
	}

	length = asn_length;
	(*objidlength)--;	/* account for expansion of first byte */

	while (length > 0 && (*objidlength)-- > 0) {
		subidentifier = 0;
		do {		/* shift and add in low order 7 bits */
			subidentifier = (subidentifier << 7) + (*(uint8_t *) bufp & ~ASN_BIT8);
			length--;
			/*
			 * Burt243241: Buffer overflow may occur if high order bit
			 * is set for every byte read. So, terminate this and return NULL
			 * incase length of OID becomes 0.
			 */
			if (length < 0) {
				return NULL;
			}
		} while (*(uint8_t *) bufp++ & ASN_BIT8);/* last byte has high bit clear */

		if (subidentifier > (uint32_t) MAX_SUBID) {
			ERROR("subidentifier too long");
			return NULL;
		}
		*oidp++ = (oid) subidentifier;
	}

	if (0 != length) {
		ERROR("OID length exceeds buffer size");
		return NULL;
	}

	/*
	 * The first two subidentifiers are encoded into the first component
	 * with the value (X * 40) + Y, where: X is the value of the first
	 * subidentifier. Y is the value of the second subidentifier.
	 */
	subidentifier = (uint32_t) objid[1];
	if (subidentifier == 0x2B) {
		objid[0] = 1;
		objid[1] = 3;
	} else {
		if (subidentifier < 40) {
			objid[0] = 0;
			objid[1] = subidentifier;
		} else if (subidentifier < 80) {
			objid[0] = 1;
			objid[1] = subidentifier - 40;
		} else {
			objid[0] = 2;
			objid[1] = subidentifier - 80;
		}
	}
	*objidlength = (int) (oidp - objid);
	DEBUGMSG(("dumpv_recv", "  ObjID: "));
	DEBUGMSGOID(("dumpv_recv", objid, *objidlength));
	DEBUGMSG(("dumpv_recv", "\n"));
	return bufp;
}

/**
 * @internal
 * asn_build_objid - Builds an ASN object identifier object containing the
 * input string.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
SK_DOMAIN_EXPORT(asn_build_objid, SK_REASON_MPSAFE);
uint8_t *
asn_build_objid(
	uint8_t         *data,	/* IN - pointer to start of object */
	size_t         *datalength,	/* IN/OUT - number of valid bytes
					 * left in buffer */
	uint8_t          type,	/* IN - ASN type of object */
	oid            *objid,	/* IN - pointer to start of input buffer */
	size_t          objidlength	/* IN - number of sub-id's in objid */
)
{
	/*
	 * ASN.1 objid ::= 0x06 asnlength subidentifier {subidentifier}*
	 * subidentifier ::= {leadingbyte}* lastbyte
	 * leadingbyte ::= 1 7bitvalue
	 * lastbyte ::= 0 7bitvalue
	 */
	size_t         asnlength;
	register oid   *op = objid;
	uint8_t          objid_size[MAX_OID_LEN];
	register uint32_t   objid_val;
	uint32_t          first_objid_val;
	register int    i;
#ifndef SNMP_NO_DEBUGGING
	uint8_t         *initdatap = data;
#endif

	/*
	 * check if there are at least 2 sub-identifiers 
	 */
	if (objidlength == 0) {
		/*
		 * there are not, so make OID have two with value of zero 
		 */
		objid_val = 0;
		objidlength = 2;
	} else if (objid[0] > 2) {
		ERROR("build objid: bad first subidentifier");
		return NULL;
	} else if (objidlength == 1) {
		/*
		 * encode the first value 
		 */
		objid_val = (op[0] * 40);
		objidlength = 2;
		op++;
	} else {
		/*
		 * combine the first two values 
		 */
		if ((op[1] > 40) &&
		    (op[0] < 2)) {
			ERROR("build objid: bad second subidentifier");
			return NULL;
		}
		objid_val = (op[0] * 40) + op[1];
		op += 2;
	}
	first_objid_val = objid_val;

	/*
	 * ditch illegal calls now 
	 */
	if (objidlength > MAX_OID_LEN) {
		return NULL;
	}

	/*
	 * calculate the number of bytes needed to store the encoded value 
	 */
	for (i = 1, asnlength = 0;;) {
		if (objid_val < (unsigned) 0x80) {
			objid_size[i] = 1;
			asnlength += 1;
		} else if (objid_val < (unsigned) 0x4000) {
			objid_size[i] = 2;
			asnlength += 2;
		} else if (objid_val < (unsigned) 0x200000) {
			objid_size[i] = 3;
			asnlength += 3;
		} else if (objid_val < (unsigned) 0x10000000) {
			objid_size[i] = 4;
			asnlength += 4;
		} else {
			objid_size[i] = 5;
			asnlength += 5;
		}
		i++;
		if (i >= (int) objidlength)
			break;
		objid_val = *op++;	/* XXX - doesn't handle 2.X (X > 40) */
	}
	
	/*
	 * store the ASN.1 tag and length 
	 */
	data = asn_build_header(data, datalength, type, asnlength);
	if (_asn_build_header_check
	    ("build objid", data, *datalength, asnlength)) {
		return NULL;
	}
	
	/*
	 * store the encoded OID value 
	 */
	for (i = 1, objid_val = first_objid_val, op = objid + 2;
	     i < (int) objidlength; i++) {
		if (i != 1) {
			objid_val = *op++;
		}
		switch (objid_size[i]) {
		case 1:
			*data++ = (uint8_t) objid_val;
			break;
			
		case 2:
			*data++ = (uint8_t) ((objid_val >> 7) | 0x80);
			*data++ = (uint8_t) (objid_val & 0x07f);
			break;
			
		case 3:
			*data++ = (uint8_t) ((objid_val >> 14) | 0x80);
			*data++ = (uint8_t) ((objid_val >> 7 & 0x7f) | 0x80);
			*data++ = (uint8_t) (objid_val & 0x07f);
			break;
			
		case 4:
			*data++ = (uint8_t) ((objid_val >> 21) | 0x80);
			*data++ = (uint8_t) ((objid_val >> 14 & 0x7f) | 0x80);
			*data++ = (uint8_t) ((objid_val >> 7 & 0x7f) | 0x80);
			*data++ = (uint8_t) (objid_val & 0x07f);
			break;
			
		case 5:
			*data++ = (uint8_t) ((objid_val >> 28) | 0x80);
			*data++ = (uint8_t) ((objid_val >> 21 & 0x7f) | 0x80);
			*data++ = (uint8_t) ((objid_val >> 14 & 0x7f) | 0x80);
			*data++ = (uint8_t) ((objid_val >> 7 & 0x7f) | 0x80);
			*data++ = (uint8_t) (objid_val & 0x07f);
			break;
		}
	}
	
	/*
	 * return the length and data ptr 
	 */
	*datalength -= asnlength;
	DEBUGDUMPSETUP("send", initdatap, data - initdatap);
	DEBUGMSG(("dumpv_send", "  ObjID: "));
	DEBUGMSGOID(("dumpv_send", objid, objidlength));
	DEBUGMSG(("dumpv_send", "\n"));
	return data;
}

/**
 * @internal
 * asn_parse_null - Interprets an ASN null type.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
uint8_t *
asn_parse_null(
	uint8_t         *data,	/* IN - pointer to start of object */
	size_t         *datalength,	/* IN/OUT - number of valid bytes
					 * left in buffer */
	uint8_t         *type	/* OUT - ASN type of object */
)
{
	/*
	 * ASN.1 null ::= 0x05 0x00
	 */
	register uint8_t *bufp = data;
	uint32_t          asn_length;

	*type = *bufp++;
	bufp = asn_parse_length(bufp, &asn_length);
	if (bufp == NULL) {
		ERROR("parse null: bad length");
		return NULL;
	}
	if (asn_length != 0) {
		ERROR("Malformed NULL");
		return NULL;
	}
	*datalength -= (bufp - data);

	DEBUGDUMPSETUP("recv", data, bufp - data);
	DEBUGMSG(("dumpv_recv", "  NULL\n"));

	return bufp + asn_length;
}


/**
 * @internal
 * asn_build_null - Builds an ASN null object.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
uint8_t *
asn_build_null(
	uint8_t         *data,	/* IN - pointer to start of object */
	size_t         *datalength,	/* IN/OUT - number of valid bytes
					 * left in buffer */
	uint8_t          type	/* IN - ASN type of object */
)
{
	/*
	 * ASN.1 null ::= 0x05 0x00
	 */
#ifndef SNMP_NO_DEBUGGING
	uint8_t         *initdatap = data;
#endif
	data = asn_build_header(data, datalength, type, 0);
	DEBUGDUMPSETUP("send", initdatap, data - initdatap);
	DEBUGMSG(("dumpv_send", "  NULL\n"));
	return data;
}

/**
 * @internal
 * asn_parse_bitstring - pulls a bitstring out of an ASN bitstring type.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  "string" is filled with the bit string.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 *
 * @param data         IN - pointer to start of object
 * @param datalength   IN/OUT - number of valid bytes left in buffer
 * @param type         OUT - asn type of object
 * @param string       IN/OUT - pointer to start of output buffer
 * @param strlength    IN/OUT - size of output buffer
 * @return Returns a pointer to the first byte past the end
 *         of this object (i.e. the start of the next object).
 *         Returns NULL on any error.
 */
uint8_t         *
asn_parse_bitstring(uint8_t * data,
                    size_t * datalength,
                    uint8_t * type, uint8_t * str, size_t * strlength)
{
    /*
     * bitstring ::= 0x03 asnlength unused {byte}*
     */
    static const char *errpre = "parse bitstring";
    register uint8_t *bufp = data;
    uint32_t          asn_length;

    *type = *bufp++;
    bufp = asn_parse_length(bufp, &asn_length);
    if (_asn_parse_length_check(errpre, bufp, data,
                                asn_length, *datalength))
        return NULL;

    if ((size_t) asn_length > *strlength) {
        _asn_length_err(errpre, (size_t) asn_length, *strlength);
        return NULL;
    }
    if (_asn_bitstring_check(errpre, asn_length, *bufp))
        return NULL;

    DEBUGDUMPSETUP("recv", data, bufp - data);
    DEBUGMSG(("dumpv_recv", "  Bitstring: "));
    DEBUGMSGHEX(("dumpv_recv", data, asn_length));
    DEBUGMSG(("dumpv_recv", "\n"));

    memmove(str, bufp, asn_length);
    *strlength = (int) asn_length;
    *datalength -= (int) asn_length + (bufp - data);
    return bufp + asn_length;
}


/**
 * @internal
 * asn_build_bitstring - Builds an ASN bit string object containing the
 * input string.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 *
 * @param data         IN - pointer to start of object
 * @param datalength   IN/OUT - number of valid bytes left in buffer
 * @param type         IN - asn type of object
 * @param string       IN - pointer to start of input buffer
 * @param strlength    IN - size of input buffer
 * @return Returns a pointer to the first byte past the end
 *         of this object (i.e. the start of the next object).
 *         Returns NULL on any error.
 */
uint8_t         *
asn_build_bitstring(uint8_t * data,
                    size_t * datalength,
                    uint8_t type, uint8_t * str, size_t strlength)
{
    /*
     * ASN.1 bit string ::= 0x03 asnlength unused {byte}*
     */
    static const char *errpre = "build bitstring";
    if (_asn_bitstring_check
        (errpre, strlength, (uint8_t)((str) ? *str :  0)))
        return NULL;

    data = asn_build_header(data, datalength, type, strlength);
    if (_asn_build_header_check(errpre, data, *datalength, strlength))
        return NULL;

    if (strlength > 0 && str)
        memmove(data, str, strlength);
    else if (strlength > 0 && !str) {
        EMS_ems_snmp_parse_error("no string passed into asn_build_bitstring");
        return NULL;
    }

    *datalength -= strlength;
    DEBUGDUMPSETUP("send", data, strlength);
    DEBUGMSG(("dumpv_send", "  Bitstring: "));
    DEBUGMSGHEX(("dumpv_send", data, strlength));
    DEBUGMSG(("dumpv_send", "\n"));
    return data + strlength;
}

/**
 * @internal
 * asn_parse_unsigned_int64 - pulls a 64 bit unsigned long out of an ASN int
 * type.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 *
 * @param data         IN - pointer to start of object
 * @param datalength   IN/OUT - number of valid bytes left in buffer
 * @param type         OUT - asn type of object
 * @param cp           IN/OUT - pointer to counter struct
 * @param countersize  IN - size of output buffer
 * @return  Returns a pointer to the first byte past the end
 *          of this object (i.e. the start of the next object).
 *          Returns NULL on any error.
 */
uint8_t         *
asn_parse_unsigned_int64(uint8_t * data,
                         size_t * datalength,
                         uint8_t * type,
                         struct counter64 * cp, size_t countersize)
{
    /*
     * ASN.1 integer ::= 0x02 asnlength byte {byte}*
     */
    static const char *errpre = "parse uint64";
    const int       uint64sizelimit = (4 * 2) + 1;
    register uint8_t *bufp = data;
    uint32_t          asn_length;
    register uint32_t low = 0, high = 0;

    if (countersize != sizeof(struct counter64)) {
        _asn_size_err(errpre, countersize, sizeof(struct counter64));
        return NULL;
    }
    *type = *bufp++;
    bufp = asn_parse_length(bufp, &asn_length);
    if (_asn_parse_length_check
        (errpre, bufp, data, asn_length, *datalength))
        return NULL;

    DEBUGDUMPSETUP("recv", data, bufp - data);
#ifdef OPAQUE_SPECIAL_TYPES
    /*
     * 64 bit counters as opaque 
     */
    if ((*type == ASN_OPAQUE) &&
        (asn_length <= ASN_OPAQUE_COUNTER64_MX_BER_LEN) &&
        (*bufp == ASN_OPAQUE_TAG1) &&
        ((*(bufp + 1) == ASN_OPAQUE_COUNTER64) ||
         (*(bufp + 1) == ASN_OPAQUE_U64))) {
        /*
         * change type to Counter64 or U64 
         */
        *type = *(bufp + 1);
        /*
         * value is encoded as special format 
         */
        bufp = asn_parse_length(bufp + 2, &asn_length);
        if (_asn_parse_length_check("parse opaque uint64", bufp, data,
                                    asn_length, *datalength))
            return NULL;
    }
#endif                          /* OPAQUE_SPECIAL_TYPES */
    if (((int) asn_length > uint64sizelimit) ||
        (((int) asn_length == uint64sizelimit) && *bufp != 0x00)) {
        _asn_length_err(errpre, (size_t) asn_length, uint64sizelimit);
        return NULL;
    }
    *datalength -= (int) asn_length + (bufp - data);
    if (*bufp & 0x80) {
        low = ~low;             /* integer is negative */
        high = ~high;
    }

    while (asn_length--) {
        high = (high << 8) | ((low & 0xFF000000) >> 24);
        low = (low << 8) | *bufp++;
    }

#if SIZEOF_LONG != 4
    if (high > 0xffffffff) {
        snmp_log(LOG_ERR,"truncating counter64 high value to 32 bits\n");
        high &= 0xffffffff;
    }
    if (low > 0xffffffff) {
        snmp_log(LOG_ERR,"truncating counter64 low value to 32 bits\n");
        low &= 0xffffffff;
    }
#endif

    cp->low = low;
    cp->high = high;

    DEBUGIF("dumpv_recv") {
        char            i64buf[I64CHARSZ + 1];
        printU64(i64buf, cp);
        DEBUGMSG(("dumpv_recv", "Counter64: ", i64buf));
    }

    return bufp;
}


/**
 * @internal
 * asn_build_unsigned_int64 - builds an ASN object containing a 64 bit integer.
 *
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 *
 * @param data         IN - pointer to start of output buffer
 * @param datalength   IN/OUT - number of valid bytes left in buffer
 * @param type         IN  - asn type of object
 * @param cp           IN - pointer to counter struct
 * @param countersize  IN - size of input buffer
 * @return  Returns a pointer to the first byte past the end
 *          of this object (i.e. the start of the next object).
 *          Returns NULL on any error.
 */
uint8_t         *
asn_build_unsigned_int64(uint8_t * data,
                         size_t * datalength,
                         uint8_t type,
                         struct counter64 * cp, size_t countersize)
{
    /*
     * ASN.1 integer ::= 0x02 asnlength byte {byte}*
     */

    register uint32_t low, high;
    register uint32_t mask, mask2;
    int             add_null_byte = 0;
    size_t          intsize;
#ifndef SNMP_NO_DEBUGGING
    uint8_t         *initdatap = data;
#endif

    if (countersize != sizeof(struct counter64)) {
        _asn_size_err("build uint64", countersize,
                      sizeof(struct counter64));
        return NULL;
    }
    intsize = 8;
    low = cp->low;
    high = cp->high;
#if SIZEOF_LONG != 4
    if (high > 0xffffffff) {
        snmp_log(LOG_ERR,"truncating counter64 high value to 32 bits\n");
        high &= 0xffffffff;
    }
    if (low > 0xffffffff) {
        snmp_log(LOG_ERR,"truncating counter64 low value to 32 bits\n");
        low &= 0xffffffff;
    }
#endif
    mask = ((uint32_t) 0xFF) << (8 * (sizeof(uint32_t) - 1));
    /*
     * mask is 0xFF000000 on a big-endian machine 
     */
    if ((uint8_t) ((high & mask) >> (8 * (sizeof(uint32_t) - 1))) & 0x80) {
        /*
         * if MSB is set 
         */
        add_null_byte = 1;
        intsize++;
    } else {
        /*
         * Truncate "unnecessary" bytes off of the most significant end of this 2's
         * complement integer.
         * There should be no sequence of 9 consecutive 1's or 0's at the most
         * significant end of the integer.
         */
        mask2 = ((uint32_t) 0x1FF) << ((8 * (sizeof(uint32_t) - 1)) - 1);
        /*
         * mask2 is 0xFF800000 on a big-endian machine 
         */
        while ((((high & mask2) == 0) || ((high & mask2) == mask2))
               && intsize > 1) {
            intsize--;
            high = (high << 8)
                | ((low & mask) >> (8 * (sizeof(uint32_t) - 1)));
            low <<= 8;
        }
    }
        data = asn_build_header(data, datalength, type, intsize);
        if (_asn_build_header_check
            ("build uint64", data, *datalength, intsize))
            return NULL;

    *datalength -= intsize;
    if (add_null_byte == 1) {
        *data++ = '\0';
        intsize--;
    }
    while (intsize--) {
        *data++ = (uint8_t) ((high & mask) >> (8 * (sizeof(uint32_t) - 1)));
        high = (high << 8)
            | ((low & mask) >> (8 * (sizeof(uint32_t) - 1)));
        low <<= 8;

    }
    DEBUGDUMPSETUP("send", initdatap, data - initdatap);
    DEBUGIF("dumpv_send") {
        char            i64buf[I64CHARSZ + 1];
        printU64(i64buf, cp);
        DEBUGMSG(("dumpv_send", i64buf));
    }
    return data;
}

/**
 * @internal
 * This function increases the size of the buffer pointed to by *pkt, which
 * is initially of size *pkt_len.  Contents are preserved **AT THE TOP END OF
 * THE BUFFER** (hence making this function useful for reverse encoding).
 * You can change the reallocation scheme, but you **MUST** guarantee to
 * allocate **AT LEAST** one extra byte.  If memory cannot be reallocated,
 * then return 0; otherwise return 1.
 *
 * @param pkt     buffer to increase
 * @param pkt_len initial buffer size
 *
 * @return 1 on success 0 on error (memory cannot be reallocated)
 */
int
asn_realloc(uint8_t ** pkt, size_t * pkt_len)
{
    if (pkt != NULL && pkt_len != NULL) {
        size_t          old_pkt_len = *pkt_len;

        DEBUGMSGTL(("asn_realloc", " old_pkt %08p, old_pkt_len %08x\n",
                    *pkt, old_pkt_len));

        if (snmp_realloc(pkt, pkt_len)) {
            DEBUGMSGTL(("asn_realloc", " new_pkt %08p, new_pkt_len %08x\n",
                        *pkt, *pkt_len));
            DEBUGMSGTL(("asn_realloc",
                        " memmove(%08p + %08x, %08p, %08x)\n", *pkt,
                        (*pkt_len - old_pkt_len), *pkt, old_pkt_len));
            memmove(*pkt + (*pkt_len - old_pkt_len), *pkt, old_pkt_len);
            memset(*pkt, (int) ' ', *pkt_len - old_pkt_len);
            return 1;
        } else {
            DEBUGMSG(("asn_realloc", " CANNOT REALLOC()\n"));
        }
    }
    return 0;

}

/**
 * @internal
 * reverse  builds an ASN header for a length with
 * length specified.
 *
 * @param pkt     IN/OUT address of the begining of the buffer.
 * @param pkt_len IN/OUT address to an integer containing the size of pkt.
 * @param offset  IN/OUT offset to the start of the buffer where to write
 * @param r       IN if not zero reallocate the buffer to fit the
 *                needed size.
 * @param length  IN - length of object
 *
 * @return 1 on success, 0 on error
 */
int
asn_realloc_rbuild_length(uint8_t ** pkt, size_t * pkt_len,
                          size_t * offset, int r, size_t length)
{
    static const char *errpre = "build length";
    char            ebuf[128];
    size_t             tmp_int;
    size_t          start_offset = *offset;

    if (length <= 0x7f) {
        if (((*pkt_len - *offset) < 1)
            && !(r && asn_realloc(pkt, pkt_len))) {
            snprintf(ebuf, sizeof(ebuf),
                    "%s: bad length < 1 :%ld, %lu", errpre,
                    (long)(*pkt_len - *offset), (unsigned long)length);
            ebuf[ sizeof(ebuf)-1 ] = 0;
            EMS_ems_snmp_parse_error(ebuf);
            return 0;
        }
        *(*pkt + *pkt_len - (++*offset)) = length;
    } else {
        while (length > 0xff) {
            if (((*pkt_len - *offset) < 1)
                && !(r && asn_realloc(pkt, pkt_len))) {
                snprintf(ebuf, sizeof(ebuf),
                        "%s: bad length < 1 :%ld, %lu", errpre,
                        (long)(*pkt_len - *offset), (unsigned long)length);
                ebuf[ sizeof(ebuf)-1 ] = 0;
                EMS_ems_snmp_parse_error(ebuf);
                return 0;
            }
            *(*pkt + *pkt_len - (++*offset)) = length & 0xff;
            length >>= 8;
        }

        while ((*pkt_len - *offset) < 2) {

            if (!(r && asn_realloc(pkt, pkt_len))) {
                snprintf(ebuf, sizeof(ebuf),
                        "%s: bad length < 1 :%ld, %lu", errpre,
                        (long)(*pkt_len - *offset), (unsigned long)length);
                ebuf[ sizeof(ebuf)-1 ] = 0;
                EMS_ems_snmp_parse_error(ebuf);
                return 0;
            }
        }

        *(*pkt + *pkt_len - (++*offset)) = length & 0xff;
        tmp_int = *offset - start_offset;
        *(*pkt + *pkt_len - (++*offset)) = tmp_int | 0x80;
    }

    return 1;
}

/**
 * @internal
 * builds an ASN header for an object with the ID and
 * length specified.
 *
 * @see asn_build_header
 *
 * @param pkt     IN/OUT address of the begining of the buffer.
 * @param pkt_len IN/OUT address to an integer containing the size of pkt.
 * @param offset  IN/OUT offset to the start of the buffer where to write
 * @param r       IN if not zero reallocate the buffer to fit the
 *                needed size.
 * @param type   IN - type of object
 * @param length   IN - length of object
 *
 * @return 1 on success, 0 on error
 */
int
asn_realloc_rbuild_header(uint8_t ** pkt, size_t * pkt_len,
                          size_t * offset, int r,
                          uint8_t type, size_t length)
{
    char            ebuf[128];

    if (asn_realloc_rbuild_length(pkt, pkt_len, offset, r, length)) {
        if (((*pkt_len - *offset) < 1)
            && !(r && asn_realloc(pkt, pkt_len))) {
            snprintf(ebuf, sizeof(ebuf),
                    "bad header length < 1 :%ld, %lu",
                    (long)(*pkt_len - *offset), (unsigned long)length);
            ebuf[ sizeof(ebuf)-1 ] = 0;
            EMS_ems_snmp_parse_error(ebuf);
            return 0;
        }
        *(*pkt + *pkt_len - (++*offset)) = type;
        return 1;
    }
    return 0;
}

/** @}*/
