/*
 *
 * Copyright (c) 1992-2006 Network Appliance, Inc.
 * All rights reserved.
 */
/*
 *****************************************************************************
 * ** md5.c -- Implementation of MD5 Message Digest Algorithm               **
 * ** Updated: 2/16/90 by Ronald L. Rivest                                  **
 * ** (C) 1990 RSA Data Security, Inc.                                      **
 * ** ************************************************************************
 */

#include <net-snmp/net-snmp-config.h>

#include <sys/types.h>
#ifdef HAVE_STRING_H
#include <legacy/string.h>
#else
#include <strings.h>
#endif
#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif

#ifdef HAVE_STDLIB_H
#include <legacy/stdlib.h>
#endif

#include <net-snmp/utilities.h>
#include <net-snmp/library/md5.h>
#include <lib/md5.h>

/*
 * SNMP_MDchecksum(data, len, SNMP_MD5): do a checksum on an arbirtrary amount of data 
 */
int
SNMP_MDchecksum(uint8_t * data, size_t len, uint8_t * mac, size_t maclen)
{
    int             rc = 0;
    MD5_CTX         md5_ctx;
    unsigned char digest[16];                         /* message digest */

    MD5InitOpt(&md5_ctx);
    while (len >= 64) {
        MD5UpdateOpt(&md5_ctx, data, 64); 
        data += 64;
        len -= 64;
    }

    MD5UpdateOpt(&md5_ctx, data, len); 

    /*
     * copy the checksum to the outgoing data (all of it that is requested). 
     */
    MD5FinalOpt(digest, &md5_ctx);
    maclen = maclen > sizeof(digest) ? sizeof(digest) : maclen;
    memcpy(mac, digest, maclen);

    return rc;
}


/*
 * SNMP_MDsign(data, len, SNMP_MD5): do a checksum on an arbirtrary amount
 * of data, and prepended with a secret in the standard fashion 
 */
int
SNMP_MDsign(uint8_t * data, size_t len, uint8_t * mac, size_t maclen,
       uint8_t * secret, size_t secretlen)
{
#define HASHKEYLEN 64

    uint8_t          K1[HASHKEYLEN];
    uint8_t          K2[HASHKEYLEN];
    uint8_t          extendedAuthKey[HASHKEYLEN];
    uint8_t          buf[HASHKEYLEN];
    unsigned char digest[16];                         /* message digest */
    size_t          i;
    uint8_t         *cp;
    int             rc = 0;
    MD5_CTX         md5_ctx;

    maclen = maclen > sizeof(digest) ? sizeof(digest) : maclen;
    /*
     * memset(K1,0,HASHKEYLEN);
     * memset(K2,0,HASHKEYLEN);
     * memset(buf,0,HASHKEYLEN);
     * memset(extendedAuthKey,0,HASHKEYLEN);
     */

    if (secretlen != 16 || secret == NULL || mac == NULL || data == NULL ||
        len <= 0 || maclen <= 0) {
        /*
         * DEBUGMSGTL(("md5","SNMP_MD5 signing not properly initialized")); 
         */
        return -1;
    }

    memset(extendedAuthKey, 0, HASHKEYLEN);
    memcpy(extendedAuthKey, secret, secretlen);
    for (i = 0; i < HASHKEYLEN; i++) {
        K1[i] = extendedAuthKey[i] ^ 0x36;
        K2[i] = extendedAuthKey[i] ^ 0x5c;
    }

    MD5InitOpt(&md5_ctx);
    MD5UpdateOpt(&md5_ctx, K1, HASHKEYLEN); 

    i = len;
    cp = data;

    while (i >= 64) {
        MD5UpdateOpt(&md5_ctx, cp, 64);
        cp += 64;
        i -= 64;
    }

    MD5UpdateOpt(&md5_ctx, cp, i); 

    memset(buf, 0, HASHKEYLEN);

    MD5FinalOpt(digest, &md5_ctx);  
    memcpy(buf, digest, sizeof(digest));
    MD5InitOpt(&md5_ctx);

    MD5UpdateOpt(&md5_ctx, K2, HASHKEYLEN);

    MD5UpdateOpt(&md5_ctx, buf, 16); 

    /*
     * copy the sign checksum to the outgoing pointer 
     */
    MD5FinalOpt(digest, &md5_ctx);
    memcpy(mac, digest, maclen);

    memset(buf, 0, HASHKEYLEN);
    memset(K1, 0, HASHKEYLEN);
    memset(K2, 0, HASHKEYLEN);
    memset(extendedAuthKey, 0, HASHKEYLEN);

    return rc;
}
