/*
 *
 * Copyright (c) 1992-2006 Network Appliance, Inc.
 * All rights reserved.
 */

/*
 * snmp_auth.c
 *
 * Community name parse/build routines.
 */
/**********************************************************************
    Copyright 1988, 1989, 1991, 1992 by Carnegie Mellon University

			 All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <net-snmp/net-snmp-config.h>


#ifdef KINETICS
#include "gw.h"
#include "fp4/cmdmacro.h"
#endif

#ifdef HAVE_STRING_H
#include <legacy/string.h>
#else
#include <strings.h>
#endif
#include <sys/types.h>
#ifdef TIME_WITH_SYS_TIME
# ifdef WIN32
#  include <sys/timeb.h>
# else
#  include <sk/time.h>
# endif
# include <legacy/time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sk/time.h>
# else
#  include <legacy/time.h>
# endif
#endif
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <net7g/netinet/sk_in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <net7g/arpa/sk_inet.h>
#endif

#ifdef HAVE_DMALLOC_H
#include <dmalloc.h>
#endif

#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif

#ifdef vms
#include <in.h>
#endif

#include <net-snmp/types.h>
#include <net-snmp/output_api.h>
#include <net-snmp/utilities.h>

#include <net-snmp/library/asn1.h>
#include <net-snmp/library/snmp_api.h>
#include <net-snmp/library/mib.h>
#include <net-snmp/library/md5.h>
#include <net-snmp/library/scapi.h>

#include <sys/param.h>
#include <sys/cdefs.h>
#include <net7g/sys/sk_socket.h>
#include <net7g/sys/sk_mbuf.h>
#include <sk/sk_domain.h>
#include <sk/sk_common.h>
#include <ontap/ems/snmp_parse_ems.h>

uint8_t *
snmp_auth_parse(uint8_t *data, size_t *length, uint8_t *sid, size_t *slen,
    int32_t *version)
{
	uint8_t          type;
        

	data = asn_parse_header(data, length, &type);
	if (data == NULL) {
		ERROR("bad header");
		return NULL;
	}
	if (type != (ASN_SEQUENCE | ASN_CONSTRUCTOR)) {
		ERROR("wrong auth header type");
		return NULL;
	}
	data = asn_parse_int(data, length, &type, version, sizeof(*version));
	if (data == NULL) {
		ERROR("bad parse of version");
		return NULL;
	}
	data = asn_parse_string(data, length, &type, sid, slen);
	if (data == NULL) {
		ERROR("bad parse of community");
		return NULL;
	}
	sid[*slen] = '\0';
	return (uint8_t *) data;
}

SK_DOMAIN_EXPORT(snmp_auth_build, SK_REASON_MPSAFE);

uint8_t *
snmp_auth_build(uint8_t *data, size_t *length, uint8_t *sid, int *slen,
    int32_t *version, int messagelen)
{
	data = asn_build_header(data, length, (uint8_t) (ASN_SEQUENCE | ASN_CONSTRUCTOR), messagelen + *slen + 5);
	if (data == NULL) {
		ERROR("buildheader");
		return NULL;
	}
	data = asn_build_int(data, length,
		     (uint8_t) (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
			     (int32_t *) version, sizeof(*version));
	if (data == NULL) {
		ERROR("buildint");
		return NULL;
	}
	data = asn_build_string(data, length,
		   (uint8_t) (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OCTET_STR),
				sid, *slen);
	if (data == NULL) {
		ERROR("buildstring");
		return NULL;
	}
	return (uint8_t *) data;
}

/*******************************************************************-o-******
 * snmp_comstr_parse
 *
 * Parameters:
 *	*data		(I)   Message.
 *	*length		(I/O) Bytes left in message.
 *	*psid		(O)   Community string.
 *	*slen		(O)   Length of community string.
 *	*version	(O)   Message version.
 *      
 * Returns:
 *	Pointer to the remainder of data.
 *
 *
 * Parse the header of a community string-based message such as that found
 * in SNMPv1 and SNMPv2c.
 */
uint8_t         *
snmp_comstr_parse(uint8_t * data,
                  size_t * length,
                  uint8_t * psid, size_t * slen, int32_t *version)
{
    uint8_t          type;
    int32_t            ver;
    size_t          origlen = *slen;

    /*
     * Message is an ASN.1 SEQUENCE.
     */
    data = asn_parse_sequence(data, length, &type,
                              (ASN_SEQUENCE | ASN_CONSTRUCTOR),
                              "auth message");
    if (data == NULL) {
        return NULL;
    }

    /*
     * First field is the version.
     */
    DEBUGDUMPHEADER("recv", "SNMP version");
    data = asn_parse_int(data, length, &type, &ver, sizeof(ver));
    DEBUGINDENTLESS();
    *version = ver;
    if (data == NULL) {
        EMS_ems_snmp_parse_error("bad parse of version");
        return NULL;
    }

    /*
     * second field is the community string for SNMPv1 & SNMPv2c 
     */
    DEBUGDUMPHEADER("recv", "community string");
    data = asn_parse_string(data, length, &type, psid, slen);
    DEBUGINDENTLESS();
    if (data == NULL) {
        EMS_ems_snmp_parse_error("bad parse of community");
        return NULL;
    }
    psid[SNMP_MIN(*slen, origlen - 1)] = '\0';
    return (uint8_t *) data;

}                               /* end snmp_comstr_parse() */




/*******************************************************************-o-******
 * snmp_comstr_build
 *
 * Parameters:
 *	*data
 *	*length
 *	*psid
 *	*slen
 *	*version
 *	 messagelen
 *      
 * Returns:
 *	Pointer into 'data' after built section.
 *
 *
 * Build the header of a community string-based message such as that found
 * in SNMPv1 and SNMPv2c.
 *
 * NOTE:	The length of the message will have to be inserted later,
 *		if not known.
 *
 * NOTE:	Version is an 'int'.  (CMU had it as a long, but was passing
 *		in a *int.  Grrr.)  Assign version to verfix and pass in
 *		that to asn_build_int instead which expects a long.  -- WH
 */
uint8_t         *
snmp_comstr_build(uint8_t * data,
                  size_t * length,
                  uint8_t * psid,
                  size_t * slen, int32_t *version, size_t messagelen)
{
    int32_t           verfix = *version;
    uint8_t         *h1 = data;
    uint8_t         *h1e;
    size_t          hlength = *length;


    /*
     * Build the the message wrapper (note length will be inserted later).
     */
    data =
        asn_build_sequence(data, length,
                           (uint8_t) (ASN_SEQUENCE | ASN_CONSTRUCTOR), 0);
    if (data == NULL) {
        return NULL;
    }
    h1e = data;


    /*
     * Store the version field.
     */
    data = asn_build_int(data, length,
                         (uint8_t) (ASN_UNIVERSAL | ASN_PRIMITIVE |
                                   ASN_INTEGER), &verfix, sizeof(verfix));
    if (data == NULL) {
        return NULL;
    }


    /*
     * Store the community string.
     */
    data = asn_build_string(data, length,
                            (uint8_t) (ASN_UNIVERSAL | ASN_PRIMITIVE |
                                      ASN_OCTET_STR), psid,
                            *(uint8_t *) slen);
    if (data == NULL) {
        return NULL;
    }


    /*
     * Insert length.
     */
    asn_build_sequence(h1, &hlength,
                       (uint8_t) (ASN_SEQUENCE | ASN_CONSTRUCTOR),
                       data - h1e + messagelen);


    return data;

}                               /* end snmp_comstr_build() */
